/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml;

import com.ssg.tools.jsonxml.common.ValidationContext;
import com.ssg.tools.jsonxml.dump.DumpFormat;
import com.ssg.tools.jsonxml.json.JSONParserLite;
import com.ssg.tools.jsonxml.json.JSONFormatterContext;
import com.ssg.tools.jsonxml.json.JSONParser;
import com.ssg.tools.jsonxml.json.JSONParserContext;
import com.ssg.tools.jsonxml.common.ParserContext;
import com.ssg.tools.jsonxml.common.FormatterContext;
import com.ssg.tools.jsonxml.json.TesterJSON;
import com.ssg.tools.jsonxml.xml.TesterXML;
import com.ssg.tools.jsonxml.xml.XMLFormatterContext;
import com.ssg.tools.jsonxml.xml.XMLParser;
import com.ssg.tools.jsonxml.xml.XMLParserContext;
import com.ssg.tools.jsonxml.common.ReflectiveBuilder;
import com.ssg.tools.jsonxml.common.StructureProcessingContext;
import com.ssg.tools.jsonxml.common.StructureProcessingException;
import com.ssg.tools.jsonxml.common.DataValidator;
import com.ssg.tools.jsonxml.json.schema.JSONSchemaException;
import com.ssg.tools.jsonxml.common.tools.URLResolver;
import com.ssg.tools.jsonxml.json.schema.JSONValidationContext;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Arrays;
import java.util.LinkedHashMap;
import java.util.Locale;
import java.util.Map;
import java.util.TimeZone;

/**
 * Factory class used to create compatible structure processor instances.
 * Useful for sharing reflective builder implementation.
 * 
 * @author ssg
 */
public class BigFactory {

    public static final String F_XML = "XML";
    public static final String F_JSON = "JSON";
    public static final String F_DUMP = "Dump";
    // factory
    private static BigFactory _instance = new BigFactory();
    // defaults == system defaults
    Locale _locale = Locale.getDefault();
    TimeZone _timeZone = TimeZone.getDefault();
    Map<String, Processors> _processors = new LinkedHashMap<String, Processors>();
    private ReflectiveBuilder _reflectiveBuilder;
    private URLResolver _urlResolver;

    /**
     * singleton ?!
     */
    private BigFactory() {
        init();
    }

    private BigFactory(Locale locale, TimeZone timeZone, ReflectiveBuilder reflectiveBuilder, URLResolver urlResolver) {
        if (locale != null) {
            _locale = locale;
        }
        if (timeZone != null) {
            _timeZone = timeZone;
        }
        if (reflectiveBuilder != null) {
            _reflectiveBuilder = reflectiveBuilder;
        }
        _urlResolver = urlResolver;
        init();
    }

    /**
     * shared factory instance access
     * @return 
     */
    public static BigFactory getInstance() {
        return _instance;
    }

    public static BigFactory getInstance(Locale locale, TimeZone timeZone, ReflectiveBuilder reflectiveBuilder) {
        return new BigFactory(locale, timeZone, reflectiveBuilder, null);
    }

    public static BigFactory getInstance(Locale locale, TimeZone timeZone, ReflectiveBuilder reflectiveBuilder, URLResolver urlResolver) {
        return new BigFactory(locale, timeZone, reflectiveBuilder, urlResolver);
    }

    public Locale getLocale() {
        return _locale;
    }

    public TimeZone getTimeZone() {
        return _timeZone;
    }

    /**
     * factory instance overrider
     * @param instance 
     */
    protected static synchronized void setInstance(BigFactory instance) {
        if (instance != null) {
            _instance = instance;
        } else {
            instance = new BigFactory();
        }
    }

    protected void init() {
        _processors.clear();
        _processors.put(F_XML, new Processors(F_XML, XMLParser.class, XMLParserContext.class, XMLFormatterContext.class, null, null, TesterXML.class));
        _processors.put(F_JSON + "Light", new Processors(F_JSON, JSONParserLite.class, JSONParserContext.class, JSONFormatterContext.class, JSONValidationContext.class, null, TesterJSON.class));
        _processors.put(F_JSON, new Processors(F_JSON, JSONParser.class, JSONParserContext.class, JSONFormatterContext.class, JSONValidationContext.class, null, TesterJSON.class));
        _processors.put(F_DUMP, new Processors(F_DUMP, null, null, DumpFormat.class, null, null, null));
    }

    public String[] getSupportedFormats() {
        String[] formats = _processors.keySet().toArray(new String[_processors.size()]);
        Arrays.sort(formats);
        return formats;
    }

    public Parser getParser(String formatName) throws StructureProcessingException {
        Processors ps = _processors.get(formatName);
        if (ps != null && ps.getParser() != null) {
            try {
                return (Parser) adjust(ps.getParser().newInstance());
            } catch (InstantiationException iex) {
            } catch (IllegalAccessException iaex) {
            }
        } else {
            if (ps == null) {
                throw new StructureProcessingException("Unknown format: " + formatName + ". Supported formats are: " + getSupportedFormats());
            }
        }
        return null;
    }

    public ParserContext getParserContext(String formatName) throws StructureProcessingException {
        return getParserContext(formatName, _locale, _timeZone);
    }

    public ParserContext getParserContext(String formatName, Locale locale, TimeZone timeZone) throws StructureProcessingException {
        Processors ps = _processors.get(formatName);
        if (ps != null && ps.getParserContext() != null) {
            try {
                Constructor[] cs = ps.getParserContext().getConstructors();
                // max: locale+time zone
                for (Constructor c : cs) {
                    Class[] prms = c.getParameterTypes();
                    if (prms.length == 2 && prms[0].equals(Locale.class) && prms[1].equals(TimeZone.class)) {
                        return (ParserContext) adjust(c.newInstance(locale, timeZone));
                    }
                }
                // middle: locale or time zone
                for (Constructor c : cs) {
                    Class[] prms = c.getParameterTypes();
                    if (prms.length == 1 && prms[0].equals(Locale.class)) {
                        return (ParserContext) adjust(c.newInstance(locale));
                    }
                    if (prms.length == 1 && prms[0].equals(TimeZone.class)) {
                        return (ParserContext) adjust(c.newInstance(timeZone));
                    }
                }
                // min: empty
                return (ParserContext) adjust(ps.getParserContext().newInstance());
            } catch (InstantiationException iex) {
            } catch (IllegalAccessException iaex) {
            } catch (InvocationTargetException itex) {
            }
        } else {
            if (ps == null) {
                throw new StructureProcessingException("Unknown format: " + formatName + ". Supported formats are: " + getSupportedFormats());
            }
        }
        return null;
    }

    public Formatter getFormatter(String formatName) throws StructureProcessingException {
        return getFormatter(formatName, _locale, _timeZone, false);
    }

    public Formatter getFormatter(String formatName, boolean indented) throws StructureProcessingException {
        return getFormatter(formatName, _locale, _timeZone, indented);
    }

    public Formatter getFormatter(String formatName, Locale locale, TimeZone timeZone, boolean indented) throws StructureProcessingException {
        Processors ps = _processors.get(formatName);
        if (ps != null && ps.getStructureFormat() != null) {
            try {
                Constructor[] cs = ps.getStructureFormat().getConstructors();
                // Max: locale+time zone+indented
                for (Constructor c : cs) {
                    Class[] prms = c.getParameterTypes();
                    if (prms.length == 3 && prms[0].equals(Locale.class) && prms[1].equals(TimeZone.class) && prms[2].equals(Boolean.class)) {
                        return (Formatter) adjust(c.newInstance(locale, timeZone, indented));
                    }
                }
                // middle: locale+time zone
                for (Constructor c : cs) {
                    Class[] prms = c.getParameterTypes();
                    if (prms.length == 2 && prms[0].equals(Locale.class) && prms[1].equals(TimeZone.class)) {
                        return (Formatter) adjust(c.newInstance(locale, timeZone));
                    }
                }
                // middle: indented or locale or time zone
                for (Constructor c : cs) {
                    Class[] prms = c.getParameterTypes();
                    if (prms.length == 1 && prms[0].equals(Boolean.class)) {
                        return (Formatter) adjust(c.newInstance(indented));
                    }
                    if (prms.length == 1 && prms[0].equals(Locale.class)) {
                        return (Formatter) adjust(c.newInstance(locale));
                    }
                    if (prms.length == 1 && prms[0].equals(TimeZone.class)) {
                        return (Formatter) adjust(c.newInstance(timeZone));
                    }
                }
                // min: empty
                return (Formatter) adjust(ps.getStructureFormat().newInstance());
            } catch (InstantiationException iex) {
            } catch (IllegalAccessException iaex) {
            } catch (InvocationTargetException itex) {
            }
        } else {
            if (ps == null) {
                throw new StructureProcessingException("Unknown format: " + formatName + ". Supported formats are: " + getSupportedFormats());
            }
        }
        return null;
    }

    public ValidationContext getValidationContext(String formatName) throws StructureProcessingException {
        return getValidationContext(formatName, _locale);
    }

    public ValidationContext getValidationContext(String formatName, Locale locale) throws StructureProcessingException {
        Processors ps = _processors.get(formatName);
        if (ps != null && ps.getValidator() != null) {
            try {
                Constructor[] cs = ps.getValidator().getConstructors();
                for (Constructor c : cs) {
                    Class[] prms = c.getParameterTypes();
                    if (prms.length == 1 && prms[0].equals(Locale.class)) {
                        return (ValidationContext) adjust(c.newInstance(locale));
                    }
                }
                // min: empty
                return (ValidationContext) adjust(ps.getValidator().newInstance());
            } catch (InstantiationException iex) {
            } catch (IllegalAccessException iaex) {
            } catch (InvocationTargetException itex) {
            }
        } else {
            if (ps == null) {
                throw new StructureProcessingException("Failed to instantiate validation context. Unknown format: " + formatName + ". Supported formats are: " + getSupportedFormats());
            }
        }
        return null;
    }

    public Tester getTester(String formatName) throws StructureProcessingException {
        Processors ps = _processors.get(formatName);
        if (ps != null && ps.getTester() != null) {
            try {
                return (Tester) adjust(ps.getTester().newInstance());
            } catch (InstantiationException iex) {
            } catch (IllegalAccessException iaex) {
            }
        } else {
            if (ps == null) {
                throw new StructureProcessingException("Failed to instantiate tester. Unknown format: " + formatName + ". Supported formats are: " + getSupportedFormats());
            }
        }
        return null;
    }

    public DataValidator getDataValidator(String formatName) throws StructureProcessingException {
        Processors ps = _processors.get(formatName);
        if (ps != null && ps.getDataValidator() != null) {
            try {
                return (DataValidator) adjust(ps.getDataValidator().newInstance());
            } catch (InstantiationException iex) {
            } catch (IllegalAccessException iaex) {
            }
        } else {
            if (ps == null) {
                throw new StructureProcessingException("Failed to instantiate data validator. Unknown format: " + formatName + ". Supported formats are: " + getSupportedFormats());
            }
        }
        return null;
    }

    public Comparer getComparer() {
        return new Comparer();
    }

    /**
     * Assigns overridable properties to created instances.
     * Create and replace with Formatter for StructureFormat instance.
     * 
     * @param instance
     * @return 
     */
    public Object adjust(Object instance) {
        if (instance instanceof StructureProcessingContext) {
            if (getReflectiveBuilder() != null) {
                ((StructureProcessingContext) instance).setReflectiveBuilder(getReflectiveBuilder());
            }
            if (instance instanceof FormatterContext) {
                instance = new Formatter((FormatterContext) instance);
            }
            if (instance instanceof ValidationContext) {
                URLResolver urlResolver = getUrlResolver(false);
                if (urlResolver != null) {
                    ((ValidationContext) instance).setUrlResolver(urlResolver);
                }
            }
        }
        return instance;
    }

    public void addFormat(String name, Processors processors) {
        if (name != null && processors != null) {
            _processors.put(name, processors);
        }
    }

    /**
     * @return the _reflectiveBuilder
     */
    protected ReflectiveBuilder getReflectiveBuilder() {
        return _reflectiveBuilder;
    }

    /**
     * @param reflectiveBuilder the _reflectiveBuilder to set
     */
    protected void setReflectiveBuilder(ReflectiveBuilder reflectiveBuilder) {
        this._reflectiveBuilder = reflectiveBuilder;
    }

    /**
     * Returns JSON schema URI resolver if any or, if enforce=true - guarantees to return it.
     * 
     * @param enforce
     * @return the _urlResolver
     */
    public URLResolver getUrlResolver(boolean enforce) {
        if (enforce) {
            try {
                return (_urlResolver != null) ? _urlResolver : new URLResolver.CacheingURLResolver();
            } catch (JSONSchemaException jsex) {
                return null;
            }
        } else {
            return _urlResolver;
        }
    }

    /**
     * @param urlResolver the _urlResolver to set
     */
    public void setUrlResolver(URLResolver urlResolver) {
        this._urlResolver = urlResolver;
    }

    public static class Processors {

        private String _formatName;
        private Class<? extends Parser> _parser;
        private Class<? extends ParserContext> _parserContext;
        private Class<? extends FormatterContext> _structureFormat;
        private Class<? extends ValidationContext> _validator;
        private Class<? extends DataValidator> _dataValidator;
        private Class<? extends Tester> _tester;

        public Processors(String name,
                Class<? extends Parser> parser,
                Class<? extends ParserContext> parserContext,
                Class<? extends FormatterContext> formatter,
                Class<? extends ValidationContext> validator,
                Class<? extends DataValidator> dataValidator,
                Class<? extends Tester> tester) {
            setFormatName(name);
            setParser(parser);
            setParserContext(parserContext);
            setStructureFormat(formatter);
            setValidator(validator);
            setDataValidator(dataValidator);
            setTester(tester);
        }

        /**
         * @return the _formatName
         */
        public String getFormatName() {
            return _formatName;
        }

        /**
         * @return the _parser
         */
        public Class<? extends Parser> getParser() {
            return _parser;
        }

        /**
         * @return the _parserContext
         */
        protected Class<? extends ParserContext> getParserContext() {
            return _parserContext;
        }

        /**
         * @return the _structureFormat
         */
        public Class<? extends FormatterContext> getStructureFormat() {
            return _structureFormat;
        }

        /**
         * @return the _validator
         */
        public Class<? extends ValidationContext> getValidator() {
            return _validator;
        }

        /**
         * @return the _tester
         */
        public Class<? extends Tester> getTester() {
            return _tester;
        }

        /**
         * @param formatName the _formatName to set
         */
        protected void setFormatName(String formatName) {
            this._formatName = formatName;
        }

        /**
         * @param parser the _parser to set
         */
        protected void setParser(Class<? extends Parser> parser) {
            this._parser = parser;
        }

        /**
         * @param parserContext the _parserContext to set
         */
        protected void setParserContext(Class<? extends ParserContext> parserContext) {
            this._parserContext = parserContext;
        }

        /**
         * @param formatter the _structureFormat to set
         */
        protected void setStructureFormat(Class<? extends FormatterContext> formatter) {
            this._structureFormat = formatter;
        }

        /**
         * @param validator the _validator to set
         */
        protected void setValidator(Class<? extends ValidationContext> validator) {
            this._validator = validator;
        }

        /**
         * @param tester the _tester to set
         */
        protected void setTester(Class<? extends Tester> tester) {
            this._tester = tester;
        }

        /**
         * @return the _dataValidator
         */
        protected Class<? extends DataValidator> getDataValidator() {
            return _dataValidator;
        }

        /**
         * @param dataValidator the _dataValidator to set
         */
        protected void setDataValidator(Class<? extends DataValidator> dataValidator) {
            this._dataValidator = dataValidator;
        }
    }

    public String getInfo() {
        StringBuilder sb = new StringBuilder();
        sb.append("BigFactory [" + getClass().getName() + "]");
        sb.append("\n  Locale             : " + getLocale());
        sb.append("\n  Time zone          : " + getTimeZone());
        sb.append("\n  reflective builder : " + getReflectiveBuilder());
        sb.append("\n  URI resolver       : " + getUrlResolver(false));
        for (String format : getSupportedFormats()) {
            Processors ps = _processors.get(format);
            sb.append("\n  format: " + format + ", data format: " + ps.getFormatName());
            sb.append("\n    parser             : " + ps._parser);
            sb.append("\n    parserContext      : " + ps._parserContext);
            sb.append("\n    formatter          : " + ps._structureFormat);
            sb.append("\n    validation context : " + ps._validator);
            sb.append("\n    data validator     : " + ps._dataValidator);
            sb.append("\n    tester             : " + ps._tester);
        }
        return sb.toString();
    }
}
