/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml.common;

import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.List;

/**
 * Designed to simplify tracking of items in object hierachy.
 * Provides set of constructure and utility methods to manipulate path.
 * 
 * Path represents namedand indexed items. Index is surrounded with "[]" and may be integer (>=0) or string value)
 * Path has depth in 2 variants: including and excluding indexes.
 * Path has various string representations:
 * - toString() - lossless
 * - toString(true) - skips indexes (i.e. parts of of path within "[]" inclusive
 * - toUTRLString() - indexes are included into path without "[]" as ordinary path elements
 * 
 * See examples of use in various structure processing context implementations.
 *
 * @author ssg
 */
public class Path {

    public static final String SEPARATOR = "/";
    public static final String ARRAY_OPEN = "[";
    public static final String ARRAY_CLOSE = "]";
    public static final int NOT_INDEXED = -1;
    public static final int NON_NUMERIC_INDEX = -2;
    protected List<String> _path = new ArrayList<String>();
    protected int _depth;
    protected int _idepth;

    private Path() {
    }

    /**
     * Creates absolute path. If null or "/" is provided creates path for root.
     * 
     * @param path 
     */
    public Path(String path) {
        add(path);
    }

    /**
     * Creates indexed path using "add(path,index)".
     * 
     * @param path
     * @param index 
     */
    public Path(String path, int index) {
        add(path, index);
    }

    /**
     * Creates path under parent with index.
     * 
     * @param parent
     * @param path
     * @param index 
     */
    public Path(Path parent, String path, int index) {
        copyPath(parent);
        add(path, index);
    }

    /**
     * Adds path item with index.
     * If index== NOT_INDEXED (-1) the no index is added.
     * 
     * @param path
     * @param index 
     */
    public void add(String path, int index) {
        add(path, (index > NOT_INDEXED) ? "" + index : null);
    }

    /**
     * Adds path with non-numeric index.
     * If index is null then non-indexd element is added.
     * @param path
     * @param index 
     */
    public void add(String path, String index) {
        if (SEPARATOR.equals(path) || path == null || path.equals("")) {
            return;
        }
        _depth++;
        _idepth++;
        if (index != null) {
            _idepth++;
        }
        _path.add(SEPARATOR);
        _path.add(path);
        if (index != null) {
            _path.add(ARRAY_OPEN);
            _path.add(index);
            _path.add(ARRAY_CLOSE);
        }
        //System.out.println("PATH: " + toString());
        //System.out.println("... added      : " + toString()+". TAG: "+path);
    }

    /**
     * Adds non-indexed element.
     * 
     * @param path 
     */
    public void add(String path) {
        add(path, NOT_INDEXED);
    }

    /**
     * Removes element. If element is index both index and element are removed.
     * 
     * @return 
     */
    public String remove() {
        //System.out.println("... remove from: " + toString());
        if (_path.size() > 0) {
            String last = null;
            if (isArray()) {
                _path.remove(_path.size() - 1);
                _path.remove(_path.size() - 1);
                _path.remove(_path.size() - 1);
                _idepth--;
            }
            if (_path.size() > 0) {
                last = _path.remove(_path.size() - 1);
                if (_path.size() > 0 && SEPARATOR.equals(_path.get(_path.size() - 1))) {
                    _path.remove(_path.size() - 1);
                }
            }
            _depth--;
            _idepth--;
            //System.out.println("... removed    : " + toString() + ". REMOVED: " + last);
            return last;
        }
        return null;
    }

    /**
     * Clears path - makes it "root".
     */
    public void clear() {
        _path.clear();
        _depth = 0;
        _idepth = 0;
    }

    @Override
    public String toString() {
        return toString(false);
    }

    /**
     * Returns string presentation of path optionally without arrays indexes (i.e. skipping "[*]").
     * 
     * @param noArrays
     * @return 
     */
    public String toString(boolean noArrays) {
        StringBuilder sb = new StringBuilder();
        int skip = 0;
        for (String s : _path) {
            if (noArrays && ARRAY_OPEN.equals(s)) {
                skip = 3;
            }
            if (skip == 0) {
                sb.append(s);
            }
            if (skip > 0) {
                skip--;
            }
        }
        return sb.toString();
    }

    /**
     * Returns path in URL-compatible mode where indexes are converted into simple path elements and
     * names and index values are URL-encoded using "UTF-8" encoding.
     * 
     * @return 
     */
    public String toURLString() {
        try {
            StringBuilder sb = new StringBuilder();
            boolean index = false;
            for (String s : _path) {
                if (SEPARATOR.equals(s)) {
                    sb.append(s);
                } else if (ARRAY_OPEN.equals(s)) {
                    sb.append('/');
                    index = true;
                } else if (ARRAY_CLOSE.equals(s)) {
                    //sb.append('/');
                    index = false;
                } else if (index) {
                    sb.append(URLEncoder.encode(s, "UTF-8"));
                } else {
                    sb.append(URLEncoder.encode(s, "UTF-8"));
                }
            }
            return sb.toString();
        } catch (UnsupportedEncodingException ueex) {
            return null;
        }
    }

    /**
     * Returns new Path object from root up to specified level (depth).
     * Indexes are not interpreted as levels.
     * 
     * @param level
     * @return 
     */
    public Path getPathForLevel(int level) {
        return getPathForLevel(level, false);
    }

    /**
     * Returns path object for specified level.
     * If level is negative or exceeds path depth then null is returned.
     */
    public Path getPathForLevel(int level, boolean indexAsLevel) {
        if (level < 0 || level > this.getDepth(indexAsLevel)) {
            return null;
        }
        Path p = new Path(null);
        p.copyPath(this);
        while (p.getDepth(indexAsLevel) > level) {
            if (p.getIndexString() != null && indexAsLevel) {
                p.setIndex(null, true);
            } else {
                p.remove();
            }
        }
        return p;
    }

    /**
     * Return parent path by cloning path and removing lat element (including index).
     * @return 
     */
    public Path getParent() {
        if (getDepth() == 0) {
            return null;
        }
        Path result = new Path();
        result.copyPath(this);
        result.remove();
        return result;
    }

    /**
     * Copies (clones) path from "from" by adding it to current path.
     * 
     * @param from 
     */
    public void copyPath(Path from) {
        for (int i = 0; i < from._path.size(); i++) {
            _path.add(from._path.get(i));
        }
        _depth += from.getDepth(false);
        _idepth += from.getDepth(true);
    }

    /**
     * Returns true if last path element is array (i.e. has index).
     * 
     * @return 
     */
    public boolean isArray() {
        if (_path.isEmpty() || _path.size() < 3) {
            return false;
        }
        if (ARRAY_CLOSE.equals(_path.get(_path.size() - 1)) && ARRAY_OPEN.equals(_path.get(_path.size() - 3))) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * Returns index as integer value or NON_INDEX or NON_NUMERIC_INDEX if none or not integer.
     * 
     * @return 
     */
    public int getIndexInt() {
        if (isArray()) {
            try {
                return Integer.parseInt(getIndexString());
            } catch (NumberFormatException nfex) {
                return NON_NUMERIC_INDEX;
            }
        } else {
            return NOT_INDEXED;
        }
    }

    /**
     * Returns string value of index or null if none.
     * 
     * @return 
     */
    public String getIndexString() {
        if (isArray()) {
            return _path.get(_path.size() - 2);
        } else {
            return null;
        }
    }

    /**
     * Assigns integer index value.
     * If current element is not index then it will be created with enforvce=true otherwise - ignored.
     * 
     * @param index
     * @param enforce 
     */
    public void setIndex(int index, boolean enforce) {
        if (index == NOT_INDEXED) {
            setIndex(null, enforce);
        } else {
            setIndex("" + index, enforce);
        }
    }

    /**
     * Assign textual index or remove index if null.
     * If current element is not index then it will be created with enforvce=true otherwise - ignored.
     * 
     * @param index
     * @param enforce 
     */
    public void setIndex(String index, boolean enforce) {
        if (isArray() || enforce) {
            if (isArray()) {
                if (index == null) {
                    _path.remove(_path.size() - 1);
                    _path.remove(_path.size() - 1);
                    _path.remove(_path.size() - 1);
                    _idepth--;
                } else {
                    _path.set(_path.size() - 2, index);
                }
            } else if (index != null) {
                _path.add(ARRAY_OPEN);
                _path.add(index);
                _path.add(ARRAY_CLOSE);
                _idepth++;
            }
        }
    }

    /**
     * Returns last path element ignoring index (if any).
     * 
     * @return 
     */
    public String getLeafName() {
        if (_path.isEmpty()) {
            return "";
        }
        if (isArray()) {
            return _path.get(_path.size() - 4);
        } else {
            return _path.get(_path.size() - 1);
        }
    }

    /**
     * Returns path depth ignoring indexes (not counting).
     * 
     * @return 
     */
    public int getDepth() {
        return getDepth(false);
    }

    /**
     * Returns depth including indexes as separate level (indexAsLevel==true) or ignoring.
     * 
     * @param indexAsLevel
     * @return 
     */
    public int getDepth(boolean indexAsLevel) {
        return (indexAsLevel) ? _idepth : _depth;
    }

    /**
     * Returns true if path is "root" (empty).
     * 
     * @return 
     */
    public boolean isRoot() {
        return _path.isEmpty();
    }

    /**
     * Creates path parsing string representation.
     * 
     * @param path
     * @return 
     */
    public static Path parse(String path) {
        Path p = new Path();
        if (path != null) {
            String[] items = path.split(SEPARATOR);
            for (String item : items) {
                if (item.endsWith(ARRAY_CLOSE) && item.indexOf(ARRAY_OPEN) > 0) {
                    int idx = item.indexOf(ARRAY_OPEN);
                    p.add(item.substring(0, idx), item.substring(idx + 1, item.length() - 1));
                } else {
                    if (item.length() > 0) {
                        p.add(item);
                    }
                }
            }
        }
        return p;
    }

    /**
     * 
     * @param parent
     * @param subpath
     * @param index
     * @return 
     */
    public static Path createFor(Path parent, String subpath, String index) {
        Path path = new Path();
        if (parent != null) {
            path.copyPath(parent);
        }
        Path sub=parse(subpath);
        sub.setIndex(index, true);
        path.copyPath(sub);
        return path;
    }
}
