/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml.common;

import com.ssg.tools.jsonxml.common.Utilities.Pair;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

/**
 * Reflective builder converts List/Map hierarchical structures into objects
 * using list of registered class mappings and reflection (getters/setters only).
 * 
 * to customize behavior on classes selection/instantiation override the following methods
 * - getReflectiveBuildClassFor: to replace evaluated class with exactly needed one
 * - getObjectInstance: to instantiate and prepare object instance when needed
 * 
 * @author ssg
 */
public class ReflectiveBuilder {

    public static String REFLECTION_COMPLEX_MAP_KEY = "RCMK_84c17cc5_64ed_4645_bd96_907ab824c653";

    /**
     * This class is used to wrap complex maps' entries.
     * @param <A>
     * @param <B> 
     */
    public static class PairForReflection<A, B> extends Pair<A, B> {

        public PairForReflection(A a, B b) {
            super(a, b);
        }

        public String getAType() {
            return (getA() != null) ? getA().getClass().getName() : null;
        }

        public String getBType() {
            return (getB() != null) ? getB().getClass().getName() : null;
        }
    }
    List<Class> _classes = new ArrayList<Class>();
    Map<String, Class> _mappedClasses = new LinkedHashMap<String, Class>();

    public ReflectiveBuilder() {
    }

    public boolean registerClass(Class clazz) {
        if (clazz != null) {
            if (_classes.contains(clazz)) {
                return true;
            }
            try {
                Constructor cs = clazz.getConstructor((Class[]) null);
                _classes.add(clazz);
                return true;
            } catch (Throwable th) {
                return false;
            }
        } else {
            return false;
        }
    }

    public boolean mapClass(String name, Class clazz) {
        if (registerClass(clazz)) {
            _mappedClasses.put(name, clazz);
            return true;
        } else {
            return false;
        }
    }

    /**
     * Override to add/remove non-reflectable classes.
     * Return false if class should not be processed reflectively.
     * Currently excludes from reflective handling all java.* and sun.* classes.
     *
     * @param object
     * @return
     */
    public boolean testIsReflectable(Object object) {
        if (object == null) {
            return false;
        }
        String className = object.getClass().getName();
        return !(className.startsWith("java.") || className.startsWith("sun."));
    }

    /**
     * Builds map of name=value pairs for object using reflection.
     * Only getters (get* without parameters) are used to do it.
     *
     * @param object
     * @return
     */
    public Map<String, Object> getReflectable(Object object) {
        Map obj2map = new LinkedHashMap();
        if (!testIsReflectable(object)) {
            // ??? we do not reflect java standard classes... -> as strings!
        } else {
            // use java reflection for arbitrary types...
            for (Pair<Object, Class> reader : this.getReflectableAccessors(object, ".+", false)) {
                Object r = reader.getA();
                String key = accessor2Key(r);
                if (key.equals("Class")) {
                    continue;
                }
                try {
                    Object v = (r instanceof Method) ? ((Method) r).invoke(object) : ((Field) r).get(object);
                    obj2map.put(key, (v != null) ? v : null);
                } catch (Throwable th) {
                    System.out.println("  extends problem: " + object);
                    th.printStackTrace(System.out);
                }
            }
        }
        return obj2map;
    }

    /**
     * Default implementation of reflective values assignment.
     * Uses getReflectableAccessors and handles Method or Field accessors only.
     * 
     * @param parent
     * @param propertyName
     * @param value
     * @throws StructureProcessingException 
     */
    public void setReflectable(Object parent, String propertyName, Object value) throws StructureProcessingException {
        Pair<Object, Class>[] writers = getReflectableAccessors(parent, propertyName, true);
        for (Pair<Object, Class> writer : writers) {
            if (writer.getB().isAssignableFrom(value.getClass())) {
                try {
                    if (writer.getA() instanceof Method) {
                        ((Method) writer.getA()).invoke(parent, value);
                    } else if (writer.getA() instanceof Field) {
                        ((Field) writer.getA()).set(parent, value);
                    } else {
                        throw new Exception("Unsupported type of accessor. Expected Method or Field, found " + writer.getA());
                    }
                    break;
                } catch (Throwable th) {
                    throw new StructureProcessingException("Failed to set " + propertyName + " for " + parent + " with " + value, th);
                }
            }
        }
    }

    /**
     * Returns expected (supported) type(s) for the property
     * Uses getters and setter only.
     * 
     * @param parent
     * @param propertyName
     * @param forWrite if true then setters are scanned otherwise - getters
     * @return 
     */
    public Pair<Object, Class>[] getReflectableAccessors(Object parent, String propertyName, boolean forWrite) {
        if (parent == null) {
            return new Pair[0];
        }
        String cacheKey = ((parent instanceof Class) ? ((Class) parent).getName() : parent.getClass().getName()) + "_" + propertyName + "_" + forWrite;
        if (_cache.containsKey(cacheKey)) {
            return _cache.get(cacheKey);
        }

        List<Pair<Object, Class>> types = new ArrayList<Pair<Object, Class>>();
        String mn = (forWrite) ? "set" + propertyName : "get" + propertyName + "|is" + propertyName;
        Method[] ms = (parent instanceof Class) ? ((Class) parent).getMethods() : parent.getClass().getMethods();
        for (Method m : ms) {
            if (m.getName().matches(mn)
                    && ((forWrite && m.getParameterTypes().length == 1) || (!forWrite && m.getParameterTypes().length == 0))) {
                types.add(new Pair<Object, Class>(m, (forWrite) ? m.getParameterTypes()[0] : m.getReturnType()));
                //m.setAccessible(true);
            }
        } // END: iterate methods

        Pair<Object, Class>[] result = types.toArray(new Pair[types.size()]);
        _cache.put(cacheKey, result);
        return result;
    }
    protected Map<String, Pair<Object, Class>[]> _cache = new HashMap<String, Pair<Object, Class>[]>();

    /**
     * Rebuild object reflectively and replace in parent.
     * 
     * @param parent
     * @param object
     * @param path
     * @param formats
     * @throws StructureProcessingException 
     */
    public void reflectiveBuild(Object parent, Object object, Path path, Formats formats, ObjectsRegistry objectsRegistry) throws StructureProcessingException {
        String name = path.getLeafName();
        Object object2 = reflectiveBuild(object, path, formats, objectsRegistry);
        //System.out.println("reflectiveBuild at "+path+": "+object+" -> "+object2);
        if (parent != null && object2 != null && !object2.equals(object)) {
            if (parent instanceof List) {
                List list = (List) parent;
                list.set(list.indexOf(object), object2);
            } else if (parent instanceof Map) {
                Map map = (Map) parent;
                map.put(name, object2);
            } else {
                setReflectable(parent, name, object2);
            }
        }
    }

    public Object reflectiveBuild(Object object, Path basePath, Formats formats, ObjectsRegistry objectsRegistry) {
        if (object == null || basePath == null) {
            return object;
        }
        Path path = new Path(null);
        path.copyPath(basePath);
        String name = path.getLeafName();
        if (_mappedClasses.containsKey(name)) {
            Object object2 = null;
            try {
                object2 = reflectiveBuild(object, _mappedClasses.get(name), path, formats, objectsRegistry);
            } catch (Throwable th) {
                th.printStackTrace(System.out);
            }
            return (object2 != null) ? object2 : object;
        } else {
            return object;
        }
    }

    /**
     * Recursively builds object for specified class from currently available context objects.
     *
     * @param context
     * @param level
     * @param name
     * @param object
     * @param clazz
     * @return
     * @throws StructureProcessingException
     */
    public Object reflectiveBuild(Object object, Class clazz, Path path, Formats formats, ObjectsRegistry objectsRegistry) throws StructureProcessingException {
        //System.out.println("reflectiveBuild: " + path + ": " + clazz + ": " + object);
        // Type adjustments section: we try to adjust value to expected value.
        if (clazz == null || object == null || !(object instanceof Map)) {
            if (Utilities.isNumericClass(clazz)) {
                Object result = formats.asNumber(object, clazz);
                if (result != null) {
                    return result;
                }
            }
            if (clazz.isArray()) {
                if (clazz.getComponentType().isEnum()) {
                    if (object instanceof String) {
                        object = string2EnumValue(clazz.getComponentType(), (String) object);
                    } else if (object instanceof Iterable) {
                        List l = new ArrayList();
                        for (Object o : (Iterable) object) {
                            if (o instanceof String) {
                                o = string2EnumValue(clazz.getComponentType(), (String) o);
                            }
                            l.add(o);
                        }
                        object = l.toArray((Object[]) java.lang.reflect.Array.newInstance(clazz.getComponentType(), l.size()));
                    }
                }
                if (!object.getClass().isArray() && !(object instanceof List)) {
                    Object[] a = (Object[]) java.lang.reflect.Array.newInstance(clazz.getComponentType(), 1);
                    object = reflectiveAdjust(null, object, clazz.getComponentType(), path, formats, objectsRegistry);
                    a[0] = object;
                    object = a;
                }
                return object;
            } else if (List.class.isAssignableFrom(clazz)) {
                return Collections.singletonList(object);
            } else {
                return object;
            }
        }
        // we have Map but expected value is List or Array - no action by default!
        if (clazz.isArray() || List.class.isAssignableFrom(clazz)) {
            return object;
        }
        Map values = (Map) object;

        // test for COMPLEX MAP i.e. <Object,Object>, not <String,Object>...
        if (isWrappedMap(values)) {
            return unwrapMap(values, path, formats, objectsRegistry);
        }

        Object result = null;
        Class itemClass = null;
        // Instantiate object. If failure then find 1st compatible
        // class in registered classes list and try it.
        try {
            result = clazz.newInstance();
            //System.out.println("  instantiated " + clazz);
        } catch (Throwable th) {
            if (clazz.equals(List.class)) {
                clazz = ArrayList.class;
            } else if (clazz.equals(Map.class)) {
                clazz = LinkedHashMap.class;
            } else {
                for (Class cl : _classes) {
                    if (clazz.isAssignableFrom(cl)) {
                        clazz = cl;
                    }
                }
            }

            // allow class override if needed
            clazz = this.getReflectiveBuildClassFor(object, clazz, path, formats, objectsRegistry);

            if (clazz != null) {
                result = getObjectInstance(object, clazz, path, values, formats, objectsRegistry);
            }
        }

        if (result == null) {
            return result;
        }


        // we got here because of normal reflection
        Set<String> processed = new HashSet<String>();
        for (Pair<Object, Class> writer : getReflectableAccessors(clazz, ".+", true)) {
            Object w = writer.getA();
            String key = accessor2Key(w);
            if (processed.contains(key)) {
                continue;
            }
            try {
                path.add(key);
                Class paramType = writer.getB();
                Object value = getValue(result, path, key, values);
                boolean valueExists = value != null;
                if (paramType.isEnum() && value instanceof String) {
                    value = string2EnumValue(paramType, (String) value);
                } else if (value != null && !paramType.isAssignableFrom(value.getClass())) {
                    value = reflectiveBuild(value, paramType, path, formats, objectsRegistry);
                } else if (value instanceof Map && ((Map) value).size() == 1 && ((Map) value).containsKey(REFLECTION_COMPLEX_MAP_KEY)) {
                    // complex map restore...
                    value = reflectiveBuild(value, paramType, path, formats, objectsRegistry);
                }
                value = reflectiveAdjust(result, value, paramType, path, formats, objectsRegistry);
                if (valueExists && value != null || !valueExists) {
                    if (w instanceof Method) {
                        // TODO: check if conditions are non-ambiguous
                        if (value == null || paramType.isAssignableFrom(value.getClass()) || paramType.isPrimitive() && Utilities.isNumericClass(value.getClass())) {
                            try {
                                ((Method) w).invoke(result, value);
                                //System.out.println(path+": "+value+", "+w);
                                processed.add(key);
                            } catch (Throwable th) {
                                //throw new StructureProcessingException("Failed to invoke method " + ((Method) r).getName() + " for " + result + " with " + value + " (" + ((value != null) ? value.getClass().getSimpleName() : "null") + "): " + th, th);
                            }
                        }
                    } else if (w instanceof Field) {
                        try {
                            ((Field) w).set(result, value);
                            processed.add(key);
                        } catch (Throwable th) {
                            throw new StructureProcessingException("Failed to set field " + ((Field) w).getName() + " for " + object + " with " + value + " (" + ((value != null) ? value.getClass().getSimpleName() : "null") + ")", th);
                        }
                    }
                }
            } finally {
                path.remove();
            }
        }

        objectsRegistry.registerProcessedObject(path.toString(), result);
        return result;
    }

    /**
     * Final or pre- decision on if any reflection is needed for object value though no default reason...
     * Converts list to Array and Array to List when needed.
     * 
     * @param object
     * @param basePath
     * @param formats
     * @param objectsRegistry
     * @return 
     */
    public Object reflectiveAdjust(Object parent, Object object, Class clazz, Path path, Formats formats, ObjectsRegistry objectsRegistry) throws StructureProcessingException {
        if (clazz.isArray() && object instanceof List) {
            object = ((List) object).toArray((Object[]) java.lang.reflect.Array.newInstance(clazz.getComponentType(), ((List) object).size()));
        }
        if (List.class.isAssignableFrom(clazz) && object != null && object.getClass().isArray()) {
            object = Arrays.asList((Object[]) object);
        }
        return object;
    }

    public Object string2EnumValue(Class enumClass, String value) {
        return PRIMITIVE_TYPE.valueOf(enumClass, value);
    }

    /**
     * Detects if map is wrapped...
     * 
     * @param values
     * @return 
     */
    public static boolean isWrappedMap(Map values) {
        return values != null && values.containsKey(REFLECTION_COMPLEX_MAP_KEY);
    }

    /**
     * Wraps map if needed to allow keys as objects (not just strings)
     * 
     * @param map
     * @return 
     */
    public static Map wrapMap(Map map) {
        Collection keys = map.keySet();
        boolean stringKeysOnly = true;
        for (Object key : keys) {
            if (!(key instanceof String)) {
                stringKeysOnly = false;
                break;
            }
        }

        if (!stringKeysOnly) {
            // proceed with EntrySet: create replacement map and update keys!!! really bad hack...
            Map newMap = new LinkedHashMap();
            List entriesList = new ArrayList();
            for (Object obj : map.entrySet()) {
                entriesList.add(new PairForReflection(((Entry) obj).getKey(), ((Entry) obj).getValue()));
            }
            newMap.put(REFLECTION_COMPLEX_MAP_KEY, entriesList);
            map = newMap;
        }

        return map;
    }

    /**
     * Performs opposite operation to wrapMap
     * 
     * @param values
     * @param path
     * @param formats
     * @return 
     */
    public Map unwrapMap(Map values, Path path, Formats formats, ObjectsRegistry objectsRegistry) {
        // test for COMPLEX MAP...
        //if (values != null && values.size() == 1 && values.containsKey(Utilities.REFLECTION_COMPLEX_MAP_KEY)) {
        if (isWrappedMap(values)) {
            // register path for wrapped map
            Path mapPath = Path.createFor(path, null, null);
            List pairs = null;
            if (true) {
                Object obj = values.get(REFLECTION_COMPLEX_MAP_KEY);
                if (obj instanceof List) {
                    pairs = (List) obj;
                } else if (obj instanceof Map && ((Map) obj).size() == 1) {
                    pairs = new ArrayList();
                    pairs.add(((Map) obj).values().iterator().next());
                } else {
                    // we failed to restore complex map.... :(
                    return values;
                }
            }

            Map result = new LinkedHashMap();
            ClassLoader classLoader = Thread.currentThread().getContextClassLoader();

            // iterate pairs and try to build nested named A and B objects
            Path p = path.createFor(path, REFLECTION_COMPLEX_MAP_KEY, null);
            int idx = 0;
            try {
                for (Object obj : pairs) {
                    p.setIndex("" + idx, true);
                    mapPath.setIndex("" + idx, true);
                    Map map = (Map) obj;
                    // cacheKey type
                    Path tmpPath = p.createFor(p, "AType", null);
                    Class classA = (objectsRegistry.isReference(tmpPath.toString(), map.get(tmpPath.getLeafName()))) ? (Class) objectsRegistry.resolveReference(tmpPath.toString(), map.get("AType")) : classLoader.loadClass((String) map.get(tmpPath.getLeafName()));
                    objectsRegistry.registerProcessedObject(tmpPath.toString(), classA);

                    // value type
                    tmpPath = p.createFor(p, "BType", null);
                    Class classB = (objectsRegistry.isReference(tmpPath.toString(), map.get(tmpPath.getLeafName()))) ? (Class) objectsRegistry.resolveReference(tmpPath.toString(), map.get("AType")) : classLoader.loadClass((String) map.get(tmpPath.getLeafName()));
                    objectsRegistry.registerProcessedObject(tmpPath.toString(), classB);

                    // cacheKey object
                    tmpPath = p.createFor(p, "A", null);
                    Object key = (objectsRegistry.isReference(tmpPath.toString(), map.get(tmpPath.getLeafName())))
                            ? objectsRegistry.resolveReference(tmpPath.toString(), map.get(tmpPath.getLeafName()))
                            : reflectiveBuild(map.get(tmpPath.getLeafName()), classA, tmpPath, formats, objectsRegistry);
                    objectsRegistry.registerProcessedObject(tmpPath.toString(), key);

                    // value object
                    tmpPath = p.createFor(p, "B", null);
                    Object val = (objectsRegistry.isReference(tmpPath.toString(), map.get(tmpPath.getLeafName())))
                            ? objectsRegistry.resolveReference(tmpPath.toString(), map.get(tmpPath.getLeafName()))
                            : reflectiveBuild(map.get(tmpPath.getLeafName()), classA, tmpPath, formats, objectsRegistry);
                    objectsRegistry.registerProcessedObject(tmpPath.toString(), val);

                    result.put(key, val);

                    idx++;
                }
            } catch (Throwable th) {
                // for any error return original object
                th.printStackTrace(System.out);
                return values;
            }
            mapPath.setIndex(null, true);
            objectsRegistry.registerProcessedObject(mapPath.toString(), result);
            return result;
        }
        return values;
    }

    /**
     * "call-back" for application-specific logic on class to use for instantiation.
     * By default returns provided expected class.
     * 
     * @param object
     * @param clazz
     * @param path
     * @param formats
     * @param objectsRegistry
     * @return
     * @throws StructureProcessingException 
     */
    Class getReflectiveBuildClassFor(Object object, Class clazz, Path path, Formats formats, ObjectsRegistry objectsRegistry) throws StructureProcessingException {
        return clazz;
    }

    /**
     * "call-back" for customized instantiation.
     * 
     * @param object
     * @param clazz
     * @param path
     * @param formats
     * @param objectsRegistry
     * @return
     * @throws StructureProcessingException 
     */
    public Object getObjectInstance(Object object, Class clazz, Path path, Map values, Formats formats, ObjectsRegistry objectsRegistry) throws StructureProcessingException {
        try {
            return clazz.newInstance();
        } catch (Throwable th) {
            if (Boolean.class.equals(clazz)) {
                Object o = values.get(path.getLeafName());
                try {
                    if (o instanceof Boolean) {
                        return new Boolean((Boolean) o);
                    } else if (o instanceof String) {
                        return Boolean.parseBoolean((String) o);
                    }
                } catch (Throwable th1) {
                }
                return null;
            }
            throw new StructureProcessingException("Failed to create instance of object " + clazz + ", " + values, th);
        }
    }

    public Object getValue(Object parent, Path path, String key, Map values) {
        Object value = values.get(key);

        if (value == null) {
            value = values.get(key.substring(0, 1).toLowerCase() + key.substring(1));
            if (value == null) {
                value = values.get(key.toUpperCase());
            }
        }
        // pre-process "null" string
        if (value != null && value.equals("null")) {
            // && !paramType.equals(String.class)) {
            value = null;
        }
        return value;
    }

    public static String accessor2Key(Object accessor) {
        if (accessor instanceof Method) {
            String s = ((Method) accessor).getName();
            if (s.startsWith("is")) {
                return s.substring(2);
            } else if (s.startsWith("get") || s.startsWith("set")) {
                return s.substring(3);
            } else {
                return s;
            }
        } else if (accessor instanceof Field) {
            return ((Field) accessor).getName();
        } else {
            return null;
        }
    }
}
