/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml.common;

import com.ssg.tools.jsonxml.common.tools.NonResizableListForArray;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.math.BigDecimal;
import java.net.URL;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Shared methods
 * 
 * @author ssg
 */
public class Utilities {

    /**
     * Binds 2 objects in single entity.
     * 
     * @param <A>
     * @param <B> 
     */
    public static class Pair<A, B> {

        private A a;
        private B b;

        public Pair(A a, B b) {
            setA(a);
            setB(b);
        }

        /**
         * @return the a
         */
        public A getA() {
            return a;
        }

        /**
         * @param a the a to set
         */
        public void setA(A a) {
            this.a = a;
        }

        /**
         * @return the b
         */
        public B getB() {
            return b;
        }

        /**
         * @param b the b to set
         */
        public void setB(B b) {
            this.b = b;
        }
    }

    /**
     * converts noun from single form into multiple as follows (samples):
     *   offer 2 offers
     *   business 2 businesses
     *   binary 2 binaries
     *
     * @param key
     * @return
     */
    public static String single2multiple(String key) {
        if (key.endsWith("s") && !key.endsWith("ss")) {
            return key;
        } else if (key.endsWith("ss")) {
            return key + "es";
        } else if (key.endsWith("y")) {
            return key.substring(0, key.length() - 1) + "ies";
        } else {
            return key + "s";
        }
    }

    /**
     * restores noun from multiple into single form reversing
     * actions applied in single2multiple as follows (samples):
     *   offers 2 offer
     *   businesses 2 business
     *   binaries 2 binary
     *
     * @param key
     * @return
     */
    public static String multiple2single(String key) {
        if (key.endsWith("ies")) {
            return key.substring(0, key.length() - 3) + 'y';
        } else if (key.endsWith("ses")) {
            return key.substring(0, key.length() - 2);
        } else if (key.endsWith("ss")) {
            return key;
        } else if (key.endsWith("s")) {
            return key.substring(0, key.length() - 1);
        } else {
            return key;
        }
    }

    /**
     * Returns true if class is numeric (byte,short,int,long,float,double or Object variant)
     * 
     * @param clazz
     * @return 
     */
    public static boolean isNumericClass(Class clazz) {
        return clazz.equals(Byte.class)
                || clazz.equals(byte.class)
                || clazz.equals(Short.class)
                || clazz.equals(short.class)
                || clazz.equals(Integer.class)
                || clazz.equals(int.class)
                || clazz.equals(Long.class)
                || clazz.equals(long.class)
                || clazz.equals(Float.class)
                || clazz.equals(float.class)
                || clazz.equals(Double.class)
                || clazz.equals(double.class)
                || clazz.equals(Number.class)
                || clazz.equals(BigDecimal.class);
    }

    /**
     * Returns true if object is of primitive type or related Object, string, enum or Date.
     * @param object
     * @return 
     */
    public static boolean isScalarOrString(Object object) {
        if (object == null) {
            return true;
        }
        Class c = object.getClass();
        if (c.isPrimitive() || c.isEnum()) {
            return true;
        }
        if (object instanceof Boolean) {
            return true;
        }
        if (object instanceof Character) {
            return true;
        }
        if (object instanceof String) {
            return true;
        }
        if (object instanceof Date) {
            return true;
        }
        if (isNumericClass(object.getClass())) {
            return true;
        }
        return false;
    }

    /**
     * Returns reader for resource.
     * 
     * @param name
     * @return
     * @throws IOException 
     */
    public static Reader getReaderForResource(String name) throws IOException {
        try {
            // works in unit tests mode
            return new InputStreamReader(Thread.currentThread().getContextClassLoader().getResourceAsStream(name), "UTF-8");
        } catch (NullPointerException npex) {
            // works in application test mode
            return new FileReader(new File(new File("./src/test/java"), name));
        }
    }

    /**
     * Recursively scans objects to get all fields and evaluates size;
     */
    public static SizeInfo estimateObjectSize(Object o, SizeInfo si) {
        if (si == null) {
            si = new SizeInfo();
        }
        if (o == null || si.isProcessed(o)) {
            return si;
        }
        //System.out.println("estimateObjectSize [" + ((si != null) ? si.level : 0) + "]: " + ((o != null) ? o.getClass().getName() : ""));// + " " + o);
        Class type = o.getClass();
        if (type.equals(String.class)) {
            si.strings++;
            if (o != null) {
                if (!si.isProcessed(o)) {
                    si.setProcessed(o);
                    si.uniqueStrings++;
                    si.bytes += ((String) o).length() * 2 + 2;
                } else {
                    si.setProcessed(o);
                }
            }
        } else if (type.isPrimitive()) {
            throw new RuntimeException("Got primitive type for field via reflection. Never got it in tests before!");
//            if (type.equals(byte.class)) {
//                si.bytes += 1;
//            } else if (type.equals(short.class)) {
//                si.bytes += 2;
//            } else if (type.equals(int.class)) {
//                si.bytes += 4;
//            } else if (type.equals(char.class)) {
//                si.bytes += 4;
//            } else if (type.equals(long.class)) {
//                si.bytes += 8;
//            } else if (type.equals(float.class)) {
//                si.bytes += 10;
//            } else if (type.equals(double.class)) {
//                si.bytes += 12;
//            } else if (type.equals(boolean.class)) {
//                si.bytes += 1;
//            }
//            si.scalars++;
//            si.setProcessed(o);
        } else if (Utilities.isScalarOrString(o)) {
            if (o instanceof Byte) {
                si.bytes += 1;
            } else if (o instanceof Short) {
                si.bytes += 2;
            } else if (o instanceof Integer) {
                si.bytes += 4;
            } else if (o instanceof Character) {
                si.bytes += 4;
            } else if (o instanceof Long) {
                si.bytes += 8;
            } else if (o instanceof Float) {
                si.bytes += 10;
            } else if (o instanceof Double) {
                si.bytes += 12;
            } else if (o instanceof Boolean) {
                si.bytes += 1;
            } else if (o instanceof Date) {
                si.bytes += 8;
            }

            si.objects++;
            si.setProcessed(o);
        } else if (type.isArray()) {
            si.arrays++;
            if (!si.isProcessed(o)) {
                si.setProcessed(o);
                si.uniqueArrays++;
                si.level++;
                for (Object vi : (List) new NonResizableListForArray(o)) {
                    estimateObjectSize(vi, si);
                }
                si.level--;
            }
        } else {
            si.objects++;
            if (!si.isProcessed(o)) {
                si.setProcessed(o);
                si.uniqueObjects++;
            } else {
                si.setProcessed(o);
            }
            if (o instanceof List) {
                si.lists++;
            }
            if (o instanceof Map) {
                si.maps++;
            }
            for (Field f : getFields(o.getClass(), null)) {
                if (si.isProcessed(f)) {
                    continue;
                }
                boolean isStatic = (f.getModifiers() & Modifier.STATIC) != 0;
                if (isStatic) {
                    si.setProcessed(f);
                    si.statics++;
                    continue;
                }

                si.fields++;
                Object value = null;
                try {
                    value = f.get(o);
                } catch (Throwable th) {
                    si.failedGets++;
                    System.out.println("Failed to get " + f.getName() + " from " + o.getClass().getName() + ": " + th);
                }
                if (value != null) {
                    estimateObjectSize(value, si);
                }
            }
        }
        return si;
    }

    /**
     * Scans recursively th class and all super-classes for declared fields.
     * 
     * @param clazz
     * @param fields
     * @return 
     */
    public static List<Field> getFields(Class clazz, List<Field> fields) {
        if (fields == null) {
            fields = new ArrayList<Field>();
        }
        if (clazz == null || clazz.equals(Object.class)) {
            return fields;
        }
        for (Field f : clazz.getDeclaredFields()) {
            f.setAccessible(true);
            fields.add(f);
        }
        return getFields(clazz.getSuperclass(), fields);
    }

    /**
     * Converts string to URL if it is represents URL or file.
     * 
     * @param s
     * @return 
     */
    public static URL asURL(String s) {
        URL result = null;
        try {
            return new URL(s);
        } catch (Throwable th) {
            try {
                File f = new File(s);
                return f.toURI().toURL();
            } catch (Throwable th1) {
            }
        }
        return null;
    }

    /**
     * Holds info on object size evaluation.
     */
    public static class SizeInfo {

        Set<Integer> _processed = new HashSet<Integer>();
        Map<Class, Integer> _counts = new LinkedHashMap<Class, Integer>();
        Map<Class, Integer> _refCounts = new LinkedHashMap<Class, Integer>();
        long objects;
        long uniqueObjects;
        long fields;
        long scalars;
        long statics;
        long strings;
        long uniqueStrings;
        long failedGets;
        long bytes;
        long arrays;
        long uniqueArrays;
        long lists;
        long maps;
        int level = 0;

        public boolean isProcessed(Object o) {
            return _processed.contains(System.identityHashCode(o));
        }

        public void setProcessed(Object o) {
            int hc = System.identityHashCode(o);
            Class c = (o != null) ? o.getClass() : null;
            if (!_processed.contains(hc)) {
                _processed.add(hc);
                if (c == null) {
                    return;
                }
                if (_counts.containsKey(c)) {
                    _counts.put(c, _counts.get(c) + 1);
                } else {
                    _counts.put(c, 1);
                }
            } else {
                if (c == null) {
                    return;
                }
                if (_refCounts.containsKey(c)) {
                    _refCounts.put(c, _refCounts.get(c) + 1);
                } else {
                    _refCounts.put(c, 1);
                }
            }
        }

        @Override
        public String toString() {
            StringBuilder sb = new StringBuilder();
            sb.append("SizeInfo: ");
            sb.append("\n     objects       : ");
            sb.append(objects);
            sb.append("\n     uniqueObjects : ");
            sb.append(uniqueObjects);
            sb.append("\n     fields        : ");
            sb.append(fields);
            sb.append("\n     scalars       : ");
            sb.append(scalars);
            sb.append("\n     statics       : ");
            sb.append(statics);
            sb.append("\n     strings       : ");
            sb.append(strings);
            sb.append("\n     uniqueStrings : ");
            sb.append(uniqueStrings);
            sb.append("\n     bytes         : ");
            sb.append(bytes);
            sb.append("\n     arrays        : ");
            sb.append(arrays);
            sb.append("\n     uniqueArrays  : ");
            sb.append(uniqueArrays);
            sb.append("\n     lists         : ");
            sb.append(lists);
            sb.append("\n     maps          : ");
            sb.append(maps);
            sb.append("\n     failedGets    : ");
            sb.append(failedGets);
            sb.append("\n  Class instances count 1sts/refs:");
            for (Class c : _counts.keySet()) {
                String s = c.getName();
                while (s.length() < 50) {
                    s += " ";
                }
                int cnt = _counts.get(c);
                if (_refCounts.containsKey(c)) {
                    int refs = _refCounts.get(c);
                    sb.append(MessageFormat.format("\n    {0} : {1,number,000000} / {2,number,000000} = {3,number,0000000}", s, cnt, refs, (cnt + refs)));
                } else {
                    sb.append(MessageFormat.format("\n    {0} : {1,number,000000}               = {2,number,0000000}", s, cnt, cnt));
                }
            }
            return sb.toString();
        }
    }
}
