/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml.common;

import com.ssg.tools.jsonxml.common.tools.NonResizableListForArray;
import com.ssg.tools.jsonxml.common.tools.URLResolver;
import java.net.URI;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Locale;
import java.util.Map;

/**
 * Provides validation context and access to ValidationSchema instance.
 * Provides utilities for simplifying data processing while validating.
 *
 * @author ssg
 */
public abstract class ValidationContext extends StructureProcessingContext {

    private List<ValidationError> _validationErrors = new ArrayList<ValidationError>();
    private URLResolver _urlResolver;

    /**
     * @return the _urlResolver
     */
    public URLResolver getUrlResolver() {
        return _urlResolver;
    }

    /**
     * @param urlResolver the _urlResolver to set
     */
    public void setUrlResolver(URLResolver urlResolver) {
        this._urlResolver = urlResolver;
    }

    /**
     * Marker interface to designate validation schema.
     */
    public static interface ValidationSchema {

        public boolean validate(Object data) throws ValidationException;

        public boolean validate(ValidationContext ctx, Object data) throws ValidationException;
    }

    public ValidationContext(Locale locale) {
        super(locale);
    }

    public ValidationContext() {
    }

    /**
     * Loads validation schema fro mspecified URI.
     * 
     * @param uri
     * @param encoding
     * @return
     * @throws ValidationException 
     */
    public abstract ValidationSchema getValidationSchema(URI uri, String encoding) throws ValidationException;

    /**
     * Performs data validation using provided schema.
     * 
     * @param schema
     * @param data
     * @return
     * @throws ValidationException 
     */
    public abstract boolean validate(ValidationSchema schema, Object data) throws ValidationException;

    public List<ValidationError> getValidationErrors() {
        return _validationErrors;
    }

    public void addValidationError(ValidationError error) {
        getValidationErrors().add(error);
    }

    public Map toMap(Object data) {
        if (data instanceof Map) {
            return (Map) data;
        } else if (getReflectiveBuilder() != null && getReflectiveBuilder().testIsReflectable(data)) {
            return getReflectiveBuilder().getReflectable(data);
        } else {
            return null;
        }
    }

    public Iterable toIterable(Object data) {
        if (data instanceof Iterable) {
            return (Iterable) data;
        } else {
            return null;
        }
    }

    public Collection toCollection(Object data) {
        if (data instanceof Collection) {
            return (Collection) data;
        } else if (data != null && data.getClass().isArray()) {
            return new NonResizableListForArray(data);
        } else {
            return null;
        }
    }

    @Override
    public void reset() {
        super.reset();
        getValidationErrors().clear();
    }
}
