/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml.common.tools;

import java.io.IOException;
import java.io.Reader;

/**
 * Implements buffered reader with unread() functionality (guarantees single character unread).
 * TODO: support declared guaranteed # of unread (to be <= buffer size) in read(buf,off,len) and unread(int)
 *
 * @author ssg
 */
public class BufferingReader extends Reader {

    public final static int EOF = '\uffff';
    Reader reader;
    char[] chars;
    int lastIdx;
    int readIdx;
    char last;

    /**
     * Default constructor. Uses internal buffer of default size: 1024 bytes.
     * 
     * @param reader 
     */
    public BufferingReader(Reader reader) {
        this.reader = reader;
        chars = new char[1024];
    }

    /**
     * Builds reader with buffer of specified size. If size is less than 64 then minimum size is used (64).
     * 
     * @param reader
     * @param bufferSize 
     */
    public BufferingReader(Reader reader, int bufferSize) {
        this.reader = reader;
        chars = new char[(bufferSize < 64) ? 64 : bufferSize];
    }

    /**
     * Creates reader wirh external buffer. 
     * If buffer is null or its size is 0 then internal buffer is used of minimal size (64).
     * 
     * @param reader
     * @param buffer 
     */
    public BufferingReader(Reader reader, char[] buffer) {
        this.reader = reader;
        chars = (buffer != null && buffer.length > 0) ? buffer : new char[64];
    }

    /**
     * Reads from reader single character. 
     * 
     * @return
     * @throws IOException 
     */
    @Override
    public int read() throws IOException {
        if (lastIdx == 0 || readIdx == lastIdx) {
            load();
        }
        if (last == EOF) {
            return EOF;
        }
//        if (readIdx < 0) {
//            readIdx = 0;
//        } else {
            last = chars[readIdx++];
//        }
        return last;
    }

    /**
     * Reads characters into buffer.
     * NOTE: do not use since bulk reads are not supporting "unread()" and vialoate it.
     * Creatred for COMPATIBILITY only FOR NOW.
     * 
     * @param cbuf
     * @param off
     * @param len
     * @return
     * @throws IOException 
     */
    @Override
    public int read(char[] cbuf, int off, int len) throws IOException {
        return reader.read(cbuf, off, len);
    }

    /**
     * Closes reader behind.
     * 
     * @throws IOException 
     */
    @Override
    public void close() throws IOException {
        reader.close();
    }

    /**
     * Loads next group of bytes.
     * 
     * @throws IOException 
     */
    void load() throws IOException {
        // save last character since it may be overriden on load
        if (lastIdx > 0) {
            last = chars[lastIdx - 1];
        }
        lastIdx = read(chars);
        if (lastIdx < chars.length) {
            if (lastIdx < 0) {
                lastIdx = 0;
            }
            chars[lastIdx++] = (char) EOF;
        }
        readIdx = 0;

    }

    /**
     * Unreads last read character.
     * 
     * @return
     * @throws IOException 
     */
    public int unread() throws IOException {
        return unread(1);
    }

    /**
     * Unreads count characters.
     * If within current buffer - just shifts index.
     * If exceeds - stores last character in "unread" variable.
     * TODO: use unread buffer.
     * 
     * @param count
     * @return
     * @throws IOException 
     */
    int unread(int count) throws IOException {
        if (readIdx < count) {
            if (readIdx == 0 && count > 0) {
                return 0;
            }
        }
        readIdx -= count;
        // fix last if unread after EOF
        if (last == EOF) {
            if (readIdx > 1) {
                readIdx--;
                last = chars[readIdx - 1];
            }

        }
        return count;
    }
}
