/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml.json;

import com.ssg.tools.jsonxml.common.Formats;
import com.ssg.tools.jsonxml.common.Path;
import com.ssg.tools.jsonxml.common.ReflectiveBuilder;
import com.ssg.tools.jsonxml.common.StructureProcessingContext;
import com.ssg.tools.jsonxml.common.Utilities;
import java.io.IOException;
import java.io.Writer;
import java.text.NumberFormat;
import java.util.Date;
import java.util.Locale;
import java.util.Map;

public class JSONFormatterLiteContext extends StructureProcessingContext {

    private static char[] SPACES = "                                        ".toCharArray();
    private Writer _writer;
    private StringBuilder _sb = new StringBuilder();
    private int _level;
    private boolean _indent;
    private boolean _wrapObjectMaps = false;
    private boolean _insertReferences = false;

    public JSONFormatterLiteContext(Writer writer) {
        setWriter(writer);
        initFormats();
    }

    /**
     * JSON uses predefined format for numbers and date/time.
     * Here we just ensure these are as specified for default handling.
     * 
     */
    public void initFormats() {
        Formats fs = getFormats();
        fs.getDateTimes().clear();
        fs.addDateTimeFormat("yyyy-MM-dd'T'HH:mm:ss'Z'");
        fs.getDates().clear();
        fs.addDateFormat("yyyy-MM-dd");
        fs.getTimes().clear();
        fs.addTimeFormat("HH:mm:ss");
        fs.getNumbers().clear();
        NumberFormat nf = NumberFormat.getInstance(Locale.US);
        nf.setGroupingUsed(false);
        fs.addNumberFormat(nf);
        setFormats(fs);
    }

    public void setWriter(Writer writer) {
        _writer = writer;
    }

    public void writeName(String name) throws IOException {
        getWriter().write('"');
        writeStringValue(name);
        getWriter().write('"');
    }

    public void writeValue(Object value) throws IOException {
        if (value == null) {
            getWriter().write("null");
        } else if (Utilities.isScalarOrString(value)) {
            if (value instanceof String) {
                writeName((String) value);
            } else if (value instanceof Boolean) {
                getWriter().write(((Boolean) value).toString());
            } else if (value instanceof Date) {
                writeName(getFormats().format(getPath(), value, 0));
            } else if (Utilities.isNumericClass(value.getClass())) {
                getWriter().write(getFormats().format(getPath(), value, 0));
            } else {
                // unsupported?
                writeComment("Unsupported value type: " + value.getClass());
            }
        } else if (isInsertReferences() && getObjectsRegistry().isProcessedObject(value)) {
            writeName(getObjectsRegistry().getProcessedObjectPath(value));
        } else {
            if (isInsertReferences()) {
                getObjectsRegistry().registerProcessedObject(getPath().toString(), value);
            }
            if (value instanceof Iterable) {
                writeArray((Iterable) value);
            } else {
                writeObject(value);
            }
        }
    }

    public void writeObject(Object object) throws IOException {
        Map map = null;
        if (object instanceof Map) {
            map = (Map) object;
        } else if (this.getReflectiveBuilder() != null && this.getReflectiveBuilder().testIsReflectable(object)) {
            map = this.getReflectiveBuilder().getReflectable(object);
        }
        if (map != null) {
            if (isWrapObjectMaps()) {
                map = ReflectiveBuilder.wrapMap(map);
            }
            getWriter().write('{');
            levelDown();
            boolean isFirst = true;
            Path path = getPath();
            for (Object key : map.keySet()) {
                if (key instanceof String) {
                    path.add((String) key);
                    if (isFirst) {
                        isFirst = false;
                    } else {
                        writeObjectSeparator();
                    }
                    writeLine();
                    writePrefix();
                    writeName((String) key);
                    writeValueSeparator();
                    writeValue(map.get(key));
                    path.remove();
                } else {
                    // TODO: wrappable or key-toString?
                    writeLine();
                    writeComment("Ignored non-string key: " + key + " in map: " + map);
                    writeLine();
                    // do not add extra ',' after comment
                    isFirst = true;
                }
            }
            levelUp();
            if (!map.isEmpty()) {
                writeLine();
                writePrefix();
            }
            getWriter().write('}');
        }
    }

    public void writeArray(Iterable array) throws IOException {
        getWriter().write('[');
        levelDown();
        boolean isFirst = true;
        int count = 0;
        Path path = getPath();
        for (Object value : array) {
            if (isFirst) {
                isFirst = false;
            } else {
                writeObjectSeparator();
            }
            writeLine();
            writePrefix();
            path.setIndex(count, true);
            writeValue(value);
            count++;
        }
        path.setIndex(null, true);
        levelUp();
        if (count > 0) {
            writeLine();
            writePrefix();
        }
        getWriter().write(']');
    }

    void levelDown() {
        setLevel(getLevel() + 1);
    }

    void levelUp() {
        setLevel(getLevel() - 1);
    }

    void writePrefix() throws IOException {
        if (isIndent()) {
            writeSpaces(getLevel() * 4);
        }
    }

    void writeValueSeparator() throws IOException {
        if (isIndent()) {
            getWriter().write(": ");
        } else {
            getWriter().write(':');
        }
    }

    void writeObjectSeparator() throws IOException {
        getWriter().write(',');
    }

    void writeLine() throws IOException {
        if (isIndent()) {
            getWriter().write('\n');
        }
    }

    void writeComment(String comment) throws IOException {
        getWriter().write("/* ");
        writeStringValue(comment);
        getWriter().write("*/");
    }

    void writeStringValue(String value) throws IOException {
        if (getSb().length() > 0) {
            getSb().delete(0, getSb().length());
        }
        for (char c : value.toCharArray()) {
            switch (c) {
                case '"':
                    getSb().append('\\');
                    getSb().append('"');
                    break;
                case '\n':
                    getSb().append('\\');
                    getSb().append('n');
                    break;
                case '\r':
                    getSb().append('\\');
                    getSb().append('r');
                    break;
                case '\t':
                    getSb().append('\\');
                    getSb().append('t');
                    break;
                case '\\':
                    getSb().append('\\');
                    getSb().append('\\');
                    break;
                case '/':
                    getSb().append('\\');
                    getSb().append(c);
                    break;
                default:
                    getSb().append(c);
            }
        }
        getWriter().write(getSb().toString());
    }

    void writeSpaces(int count) throws IOException {
        if (count <= 0) {
            return;
        }
        while (count > SPACES.length) {
            getWriter().write(SPACES);
            count -= SPACES.length;
        }
        if (count < SPACES.length) {
            getWriter().write(SPACES, 0, count);
        }
    }

    /**
     * @return the _writer
     */
    protected Writer getWriter() {
        return _writer;
    }

    /**
     * @return the _sb
     */
    protected StringBuilder getSb() {
        return _sb;
    }

    /**
     * @param sb the _sb to set
     */
    protected void setSb(StringBuilder sb) {
        this._sb = sb;
    }

    /**
     * @return the _level
     */
    protected int getLevel() {
        return _level;
    }

    /**
     * @param level the _level to set
     */
    protected void setLevel(int level) {
        this._level = level;
    }

    /**
     * @return the _indent
     */
    protected boolean isIndent() {
        return _indent;
    }

    /**
     * @param indent the _indent to set
     */
    protected void setIndent(boolean indent) {
        this._indent = indent;
    }

    /**
     * @return the _wrapObjectMaps
     */
    protected boolean isWrapObjectMaps() {
        return _wrapObjectMaps;
    }

    /**
     * @param wrapObjectMaps the _wrapObjectMaps to set
     */
    protected void setWrapObjectMaps(boolean wrapObjectMaps) {
        this._wrapObjectMaps = wrapObjectMaps;
    }

    /**
     * Returns true if insertReferences is set to true AND objects registry is defined.
     * @return the _insertReferences
     */
    protected boolean isInsertReferences() {
        return _insertReferences && getObjectsRegistry() != null;
    }

    /**
     * @param insertReferences the _insertReferences to set
     */
    protected void setInsertReferences(boolean insertReferences) {
        this._insertReferences = insertReferences;
    }
}
