/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml.json;

import com.ssg.tools.jsonxml.Parser;
import com.ssg.tools.jsonxml.common.ParserContext;
import com.ssg.tools.jsonxml.common.Formats;
import com.ssg.tools.jsonxml.common.StructureErrorInfo;
import com.ssg.tools.jsonxml.common.StructureProcessingException;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.Reader;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

/**
 * Implements parsing of JSON-encoded texts in accordance with JSON specification.
 * Supports strict and non-strict modes. In strict mode all names must be quoted,
 * non-strict mode allows non-quoted names.
 * 
 * Keeps track of processing using StructureProcessingContext allowing 
 * deserialization of JSON-unfriendly structures (like Map<Object,Object>)
 * and object references.
 * 
 * @author ssg
 */
public class JSONParser extends Parser {

    static Logger log = Logger.getLogger(Parser.class.getName());

    public Object postProcessStringValue(JSONParserContext context, String object) {
        return object;
    }

    public Object postProcessObject(JSONParserContext context, Object object, boolean resolveReferences) {
        //System.out.println("postProcessObject: "+context.getPath()+"; "+((object!=null) ? object.getClass().getSimpleName() : "null"));
        Object result = object;
        if (result instanceof String) {
            result = postProcessStringValue(context, (String) result);
            if (result instanceof String && resolveReferences) {
                result = context.getObjectsRegistry().resolveReference(context.getPath().toString(), result);
            }
        }
        return result;
    }

    public Object parse(Reader r) throws IOException {
        return parse(r, (Formats) null);
    }

    public Object parse(Reader r, Formats formats) throws IOException {
        JSONParserContext context = new JSONParserContext();
        if (formats != null) {
            context.setFormats(formats);
        }
        return parse(r, context);
    }

    public Object parse(Reader r, ParserContext jsonContext) throws IOException {
        if (!r.markSupported()) {
            r = new BufferedReader(r);
        }

        final JSONParserContext context = (jsonContext != null)
                ? ((jsonContext instanceof JSONParserContext))
                ? (JSONParserContext) jsonContext
                : new JSONParserContext(jsonContext)
                : new JSONParserContext();

        // "inject" line/position tracker for diagnosis...
        r = new BufferedReader(r) {

            int buf_limit = 0;
            int line = 1;
            long pos = 1;
            long globalPos = 0;
            int old_line;
            long old_pos;
            int last_ch = 0;

            @Override
            public void mark(int readAheadLimit) throws IOException {
                super.mark(readAheadLimit);
                buf_limit = readAheadLimit;
                old_line = line;
                old_pos = pos;
            }

            @Override
            public int read() throws IOException {
                int ch = super.read();
                if (ch != READER_EOF) {
                    globalPos++;
                    pos++;
                    if (ch == '\r' || ch == '\n') {
                        pos = 1;
                        if (last_ch != ch && (last_ch == '\n' || last_ch == '\r')) {
                            line++;
                        }
                    }
                    last_ch = ch;
                }
                copyToContext();
                return ch;
            }

            @Override
            public void reset() throws IOException {
                super.reset();
                globalPos -= buf_limit;
                line = old_line;
                pos = old_pos;
                copyToContext();
            }

            void copyToContext() {
                context.setLine(line);
                context.setPosition(pos);
                context.setGlobalPosition(globalPos);
            }
        };

        return parseJSON(r, false, context);
    }

    /**
     * special notes on context._path usage:
     *   - added for maps.
     *   - for created list existing name is enforced to be array.
     * @param r
     * @param asName
     * @param context
     * @return
     * @throws IOException
     */
    Object parseJSON(Reader r, boolean asName, JSONParserContext context) throws IOException {
        StringBuilder sb = new StringBuilder();
        Object result = null;
        char ch = nextNonSpaceChar(r);
        while (ch == '/') {
            skipComment(r);
            ch = nextNonSpaceChar(r);
        }
        switch (ch) {
            case READER_EOF:
                context.setEOF(true);
                break;
            case JSONParserContext.OPEN_LIST:
                result = new ArrayList();
                context.getStack().push(result);
                if (1 == 1) {
                    // ensure path is ARRAY now
                    context.getPath().setIndex(0, true);
                }
                // iterate...
                if (result != null) {
                    Object value = parseJSON(r, false, context);
                    while (!result.equals(value) && !context.isEOF()) {
                        if (!JSONParserContext.OBJECT_DELIMITER_TAG.equals(value)) {
                            ((List) result).add(value);
                        }
                        context.getPath().setIndex(((List) result).size(), true);
                        value = parseJSON(r, false, context);
                    }
                    if (context.isEOF()) {
                        //System.out.println("parseJSON: ABNORMAL LIST END @"+context.getPath());
                        context.addError(new StructureErrorInfo(" ABNORMAL LIST END @" + context.getPath(), "" + context.CLOSE_LIST, "EOF"));
                    }
                }
                break;
            case JSONParserContext.CLOSE_LIST:
                if (context.getStack().peek() instanceof List) {
                    result = context.getStack().pop();
                    context.getPath().setIndex(null, true);
                    //context._path.remove();
                } else {
                    StructureErrorInfo error = new StructureErrorInfo(context.getLine(), (int) context.getPosition(), "'" + context.getJsonFormat().getObjectSeparator() + "' or '}'", "'" + ch + "'");
                    context.addError(error);
                    throw new StructureProcessingException("No open list to close.", null, error);
                }
                break;
            case JSONParserContext.OPEN_OBJECT:
                result = new LinkedHashMap();
                context.getStack().push(result);
                // iterate...
                if (result != null) {
                    Object name = parseJSON(r, true, context);
                    Object value = null;
                    if (!result.equals(name)) {
                        context.getPath().add(name.toString());
                        value = parseJSON(r, false, context);
                        context.getPath().remove();
                    }
                    while (name != null && !result.equals(name) && !context.isEOF()) {
                        if (!JSONParserContext.OBJECT_DELIMITER_TAG.equals(name)) {
                            ((Map) result).put(name, value);
                        }
                        name = parseJSON(r, true, context);
                        if (name != null && !result.equals(name) && !JSONParserContext.OBJECT_DELIMITER_TAG.equals(name)) {
                            context.getPath().add(name.toString());
                            value = parseJSON(r, false, context);
                            context.getPath().remove();
                        } else {
                            value = null;
                        }
                    }
                    if (context.isEOF()) {
                        //System.out.println("parseJSON: ABNORMAL OBJECT END @"+context.getPath());
                        context.addError(new StructureErrorInfo("ABNORMAL OBJECT END @" + context.getPath(), "" + context.CLOSE_OBJECT, "EOF"));
                    }
                }
                break;
            case JSONParserContext.CLOSE_OBJECT:
                if (context.getStack().peek() instanceof Map) {
                    result = context.getStack().pop();
                } else {
                    StructureErrorInfo error = new StructureErrorInfo(context.getLine(), (int) context.getPosition(), "'" + context.getJsonFormat().getObjectSeparator() + "' or ']'", "'" + ch + "'");
                    context.addError(error);
                    throw new StructureProcessingException("No open object to close.", null, error);
                }
                break;
            case JSONParserContext.OBJECTS_DELIMITER:
                result = JSONParserContext.OBJECT_DELIMITER_TAG;
                break;
            default:
                if (asName) {
                    String name = null;
                    if (ch == '\'' || ch == '"') {
                        name = readQuotedString(r, ch);
                        readQuotedString(r, JSONParserContext.VALUE_DELIMITER);
                    } else {
                        if (context.isStrict()) {
                            StructureErrorInfo error = new StructureErrorInfo(context.getLine(), (int) context.getPosition(), "'" + context.getJsonFormat().getQuoteName() + "'", "'" + ch + "'");
                            context.addError(error);
                            throw new StructureProcessingException("Illegal start of identifier name", null, error);
                        } else {
                            name = readQuotedString(r, JSONParserContext.VALUE_DELIMITER);
                        }
                    }
                    result = name;
                } else {
                    if (ch == '\'' || ch == '"') {
                        result = readQuotedString(r, ch);
                        result = context.getJsonFormat().decodeStringValue((String) result);
                    } else {
                        String s = ch + readNotString(r);
                        if (s.equals("null")) {
                            return null;
                        } else {
                            Object primitive = context.getFormats().toPrimitiveValue(s);
                            if (primitive != null) {
                                result = primitive;
                            } else {
                                result = context.getJsonFormat().decodeStringValue(s);//s;
                            }
                        }
                    }
                }
                break;
        }
        if (!asName) {
            result = postProcessObject(context, result, true);
        }
        context.getObjectsRegistry().registerProcessedObject(context.getPath().toString(), result);
        return result;
    }

    String readQuotedString(Reader r, char quoteChar) throws IOException {
        StringBuilder sb = new StringBuilder();
        char ch = readChar(r);
        // TODO: check for escaped quote char to avoid false string ending
        while (ch != READER_EOF && ch != quoteChar) {
            sb.append(ch);
            ch = readChar(r);
        }
        return sb.toString();
    }

    String readNotString(Reader r) throws IOException {
        StringBuilder sb = new StringBuilder();
        r.mark(1);
        char ch = readChar(r);
        while (ch != READER_EOF && !isWhiteSpace(ch) && ch != JSONParserContext.CLOSE_OBJECT && ch != JSONParserContext.CLOSE_LIST && ch != JSONParserContext.OBJECTS_DELIMITER) {
            sb.append(ch);
            r.mark(1);
            ch = readChar(r);
        }
        if (ch == JSONParserContext.CLOSE_OBJECT || ch == JSONParserContext.CLOSE_LIST || ch == JSONParserContext.OBJECTS_DELIMITER) {
            r.reset();
        }
        return sb.toString();
    }

    void skipComment(Reader r) throws IOException {
        int ch = r.read();
        if (ch == '/') {
            ch = r.read();
            while (ch != READER_EOF) {
                ch = r.read();
                if (ch == '\n' || ch == '\r') {
                    return;
                }
            }
        } else if (ch == '*') {
            int last = r.read();
            ch = r.read();
            while (ch != READER_EOF) {
                last = ch;
                ch = r.read();
                if (ch == '/' && last == '*') {
                    return;
                }
            }
        } else {
            throw new IOException("Unrecognized comment type.");
        }
    }
}
