/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml.json;

import com.ssg.tools.jsonxml.common.tools.BufferingReader;
import com.ssg.tools.jsonxml.Parser;
import com.ssg.tools.jsonxml.common.Formats;
import java.io.IOException;
import java.io.Reader;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

/**
 * Optimized JSON parser.
 * 
 * Uses context to store source (reader) and result (object).
 * Parser is implemented as instance to allow easy override for created Object and/or List.
 * By default uses LinkedHashMap and ArrayList instances.
 * 
 * @author ssg
 */
public class JSONParserLite extends Parser {

    public static final int EOF = BufferingReader.EOF;

    @Override
    public Object parse(Reader r) throws IOException {
        JSONParserLiteContext ctx = new JSONParserLiteContext(r);
        parse(ctx);
        return ctx.getObject();
    }

    @Override
    public Object parse(Reader r, Formats formatters) throws IOException {
        // TODO: Parser compatibility : utilize formatters or ignore?
        return parse(r);
    }

    @Override
    public Object parse(Reader r, com.ssg.tools.jsonxml.common.ParserContext context) throws IOException {
        // TODO: Parser compatibility : utilize parser context... or ignore?
        return parse(r);
    }

    public enum NEXT {

        ITEM, NAME, VALUE_SEPARATOR, EOO, EOL, EOF
    }

    public JSONParserLite() {
    }

    /**
     * Use this method to parse reader.
     * NOTE: If parsing context or its configuration is needed then
     * use class instance "parse" method.
     * 
     * @param reader
     * @return
     * @throws IOException 
     */
    public static Object parse(BufferingReader reader) throws IOException {
        JSONParserLite tjp = new JSONParserLite();
        JSONParserLiteContext ctx = new JSONParserLiteContext(reader);
        tjp.parse(ctx);
        return ctx.getObject();
    }

    /**
     * Parses data presented via ParserContext into context's object.
     * 
     * @param ctx
     * @throws IOException
     * @throws Exception 
     */
    public void parse(JSONParserLiteContext ctx) throws IOException {
        int c;
        boolean objectSeparatorDetected = false;
        while ((c = ctx.read()) != EOF) {
            switch (c) {
                case ' ':
                case '\t':
                case '\n':
                case '\r':
                    break;
                case '"':
                    ctx.object = readString(ctx.reader, ctx.string);
                    if (ctx.next == NEXT.NAME) {
                        ctx.next = NEXT.VALUE_SEPARATOR;
                    } else {
                        ctx.next = NEXT.ITEM;
                    }
                    return;
                case '{':
                    if (ctx.next != NEXT.ITEM) {
                        throw new IOException("Syntax error: found OBJECT, expected " + ctx.next);
                    } {
                    Map map = newObject();
                    ctx.next = NEXT.NAME;
                    ctx.inObject = true;
                    parse(ctx);
                    while (ctx.next != NEXT.EOO) {
                        if (ctx.next == NEXT.EOF) {
                            throw new IOException("EOF detected but object is not closed.");
                        }
                        String key = (String) ctx.object;
                        parse(ctx);
                        if (ctx.next != NEXT.ITEM) {
                            throw new IOException("Next item to read after value separator must be " + NEXT.ITEM + ", found " + ctx.next);
                        }
                        parse(ctx);
                        if (map != null) {
                            map.put(key, ctx.object);
                        }
                        ctx.object = null;
                        ctx.next = NEXT.NAME;
                        ctx.inObject = true;
                        parse(ctx);
                    }
                    ctx.object = map;
                    ctx.next = NEXT.ITEM;
                    return;
                }
                case '}':
                    if (objectSeparatorDetected) {
                        throw new IOException("Syntax error: no value OBJECT DELIMITER (','), found '" + (char) c + "'");
                    }
                    ctx.next = NEXT.EOO;
                    return;
                case '[':
                    if (ctx.next != NEXT.ITEM) {
                        throw new IOException("Syntax error: found LIST, expected " + this);
                    } {
                    List list = newList();
                    ctx.next = NEXT.ITEM;
                    ctx.inObject = false;
                    parse(ctx);
                    while (ctx.next != NEXT.EOL) {
                        if (ctx.next == NEXT.EOF) {
                            throw new IOException("EOF detected but array is not closed.");
                        }
                        if (list != null) {
                            list.add(ctx.object);
                        }
                        ctx.object = null;
                        ctx.next = NEXT.ITEM;
                        ctx.inObject = false;
                        parse(ctx);
                    }
                    ctx.object = list;
                    ctx.next = NEXT.ITEM;
                    return;
                }
                case ']':
                    if (objectSeparatorDetected) {
                        throw new IOException("Syntax error: no value OBJECT DELIMITER (','), found '" + (char) c + "'");
                    }
                    ctx.next = NEXT.EOL;
                    return;
                case ',':
                    if (objectSeparatorDetected) {
                        throw new IOException("Syntax error: found duplicate OBJECT DELIMITER (','), expected " + ctx.next);
                    } else {
                        objectSeparatorDetected = true;
                    }
                    if (ctx.inObject) {
                        ctx.next = NEXT.NAME;
                    } else {
                        ctx.next = NEXT.ITEM;
                    }
                    break;
                case ':':
                    if (ctx.next != NEXT.VALUE_SEPARATOR) {
                        throw new IOException("Syntax error: found VALUE DELIMITER, expected " + ctx.next);
                    }
                    ctx.next = NEXT.ITEM;
                    return;
                case '/':
                    c = ctx.read();
                    if (c == '*') {
                        int last = c;
                        while ((c = ctx.read()) != EOF) {
                            if (c == '/' && last == '*') {
                                break;
                            }
                            last = c;
                        }
                    } else if (c == '/') {
                        // skip line comment
                        while ((c = ctx.read()) != EOF) {
                            if (c == '\n' || c == '\r') {
                                break;
                            }
                        }
                    } else {
                        throw new IOException("Unexpected characters /" + (char) c);
                    }
                    break;
                default:
                    if (ctx.next != NEXT.ITEM) {
                        throw new IOException("Syntax error: found non-string VALUE, expected " + ctx.next);
                    }
                    ctx.reader.unread();
                    ctx.object = readValue(ctx.reader, ctx.string);
                    return;
            }
        }
        ctx.next = NEXT.EOF;
    }

    /**
     * Reads string into provided StringBuilder.
     * 
     * @param r
     * @param sb
     * @return
     * @throws IOException 
     */
    public String readString(Reader r, StringBuilder sb) throws IOException {
        {
            int len = sb.length();
            if (len > 0) {
                sb.delete(0, len);
            }
        }
        int c;
        while ((c = r.read()) != '"') {
            switch (c) {
                case '\\':
                    c = r.read();
                    switch (c) {
                        case 'r':
                            sb.append('\r');
                            break;
                        case 'n':
                            sb.append('\n');
                            break;
                        case 't':
                            sb.append('\t');
                            break;
                        case '\\':
                            sb.append('\\');
                            break;
                        case '/':
                            sb.append('/');
                            break;
                        case '"':
                            sb.append('"');
                            break;
                        default:
                            throw new IOException("Unknown escape character: " + c);
                    }
                    break;
                case 65535:
                case -1:
                    throw new IOException("Unexpected EOF while reading STRING");
                default:
                    sb.append((char) c);
            }
        }
        return sb.toString();
    }

    /**
     * Reads string into character array and builds string from result.
     * NOTE: string size is limited to buffer size.
     * 
     * @param r
     * @param buffer
     * @return
     * @throws IOException 
     */
    public static String readString(Reader r, char[] buffer) throws IOException {
        int c;
        int charIdx = 0;
        while ((c = r.read()) != '"') {
            switch (c) {
                case '\\':
                    c = r.read();
                    switch (c) {
                        case 'r':
                            buffer[charIdx++] = '\r';
                            break;
                        case 'n':
                            buffer[charIdx++] = '\n';
                            break;
                        case 't':
                            buffer[charIdx++] = '\t';
                            break;
                        case '\\':
                            buffer[charIdx++] = '\\';
                            break;
                        case '/':
                            buffer[charIdx++] = '/';
                            break;
                        case '"':
                            buffer[charIdx++] = '"';
                            break;
                        default:
                            throw new IOException("Unknown escape character: " + c);
                    }
                    break;
                case 65535:
                case -1:
                    throw new IOException("Unexpected EOF while reading STRING");
                default:
                    buffer[charIdx++] = (char) c;
            }
        }
        return new String(buffer, 0, charIdx);
    }

    /**
     * Reads value into StringBuilder and then interprets it as number, boolean, or null.
     *
     * @param r
     * @param sb
     * @return
     * @throws IOException 
     */
    public static Object readValue(BufferingReader r, StringBuilder sb) throws IOException {
        boolean isInt = true;
        boolean isDouble = true;

        {
            int len = sb.length();
            if (len > 0) {
                sb.delete(0, len);
            }
        }

        int c;
        while ((c = r.read()) != EOF) {
            switch (c) {
                case ',':
                case '}':
                case ']':
                case '\n':
                case '\t':
                case '\r':
                case ' ':
                case '/':
                    String v = sb.toString();
                    if (isInt) {
                        try {
                            //System.out.println("as Int: "+sb.toString());
                            r.unread();
                            return Integer.parseInt(v);
                        } catch (Exception ex) {
                            throw new IOException("Invalid integer: " + ex);
                        }
                    } else if (isDouble) {
                        try {
                            //System.out.println("as Double: "+sb.toString());
                            r.unread();
                            return Double.parseDouble(v);
                        } catch (Exception ex) {
                            throw new IOException("Invalid double: " + ex);
                        }
                    } else {
                        r.unread();
                        if ("true".equals(v)) {
                            return true;
                        } else if ("false".equals(v)) {
                            return false;
                        } else if ("null".equals(v)) {
                            return null;
                        } else {
                            throw new IOException("Unparseable value: " + v);
                        }
                    }
                case '\"':
                case '\\':
                case ':':
                case '[':
                case '{':
                    throw new IOException("Unexpected character in/after value: " + c + ": " + (char) c);
                default:
                    sb.append((char) c);
                    if (isInt) {
                        if ((c == '-' && sb.length() == 1) || Character.isDigit(c)) {
                        } else {
                            isInt = false;
                        }
                    }
                    if (isDouble) {
                        if (!isInt) {
                            switch (c) {
                                case '.':
                                case 'e':
                                case 'E':
                                case '-':
                                    break;
                                default:
                                    if (!Character.isDigit(c)) {
                                        isDouble = false;
                                    }
                            }
                        }
                    }
            }
        }
        throw new IOException("Unexpected EOF");
    }

    /**
     * Reads value into character array and then interprets it as number, boolean, or null.
     * NOTE: string size is limited to buffer size.
     * 
     * @param r
     * @param buffer
     * @return
     * @throws IOException 
     */
    public static Object readValue(BufferingReader r, char[] buffer) throws IOException {
        boolean isInt = true;
        boolean isDouble = true;

        int charIdx = 0;

        int c;
        while ((c = r.read()) != EOF) {
            switch (c) {
                case ',':
                case '}':
                case ']':
                case '\n':
                case '\t':
                case '\r':
                case ' ':
                case '/':
                    String v = new String(buffer, 0, charIdx);
                    if (isInt) {
                        try {
                            r.unread();
                            return Integer.parseInt(v);
                        } catch (Exception ex) {
                            throw new IOException("Invalid integer: " + ex);
                        }
                    } else if (isDouble) {
                        try {
                            r.unread();
                            return Double.parseDouble(v);
                        } catch (Exception ex) {
                            throw new IOException("Invalid double: " + ex);
                        }
                    } else {
                        r.unread();
                        if ("true".equals(v)) {
                            return true;
                        } else if ("false".equals(v)) {
                            return false;
                        } else if ("null".equals(v)) {
                            return null;
                        } else {
                            throw new IOException("Unparseable value: " + v);
                        }
                    }
                case '\"':
                case '\\':
                case ':':
                case '[':
                case '{':
                    throw new IOException("Unexpected character in/after value: " + c + ": " + (char) c);
                default:
                    //ctx.sb.append((char) c);
                    buffer[charIdx++] = (char) c;
                    if (isInt) {
                        if ((c == '-' && charIdx == 1) || Character.isDigit(c)) {
                        } else {
                            isInt = false;
                        }
                    }
                    if (isDouble) {
                        if (!isInt) {
                            switch (c) {
                                case '.':
                                case 'e':
                                case 'E':
                                case '-':
                                    break;
                                default:
                                    if (!Character.isDigit(c)) {
                                        isDouble = false;
                                    }
                            }
                        }
                    }
            }
        }
        throw new IOException("Unexpected EOF");
    }

    /**
     * Overridable method for new Object instance creation.
     * E.g. it might be some JSONObject class instance.
     * Default is LinkedHashMap.
     * 
     * @return 
     */
    public Map newObject() {
        return new LinkedHashMap();
    }

    /**
     * Overridable method for new Array instance creation.
     * Default is ArrayList.
     * 
     * @return 
     */
    public List newList() {
        return new ArrayList();
    }
}
