/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml.json;

import com.ssg.tools.jsonxml.Parser;
import com.ssg.tools.jsonxml.Tester;
import com.ssg.tools.jsonxml.common.CRITICALITY;
import com.ssg.tools.jsonxml.common.ValidationError;
import com.ssg.tools.jsonxml.json.schema.JSONSchema;
import com.ssg.tools.jsonxml.json.schema.JSONValidationContext;
import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import javax.xml.transform.stream.StreamSource;

/**
 * Detect if URL points to JSON text.
 * File is assumed to be JSON if 1st character (after any comments and/or white spaces)
 * is either '{' (object) or '[' (array).
 * Skipped comments are:
 *   - inline (start with '//')
 *   - block ( start with '/*', end with reverse sequence AFTER start sequence, not nested)
 * @author ssg
 */
public class TesterJSON extends Tester {

    @Override
    public boolean test(URL source) throws IOException {
        InputStream is = null;
        try {
            is = new BufferedInputStream(source.openStream(), 256);
            int c = 0;
            while ((c = is.read()) != Parser.READER_EOF) {
                if (Parser.isWhiteSpace((char) c)) {
                    continue;
                }
                if (c == '/') {
                    // skip comment(s)
                    c = is.read();
                    if (c == '/') {
                        // line comment
                        boolean OK = false;
                        while ((c = is.read()) != Parser.READER_EOF) {
                            if (c == '\n' || c == '\r') {
                                OK = true;
                                break;
                            }
                        }
                        if (OK) {
                            continue;
                        } else {
                            return false;
                        }
                    } else if (c == '*') {
                        // block comment
                        boolean OK = false;
                        int last = 0;
                        while ((c = is.read()) != Parser.READER_EOF) {
                            if (c == '/' && last == '*') {
                                OK = true;
                                break;
                            }
                            last = c;
                        }
                        if (OK) {
                            continue;
                        } else {
                            return false;
                        }
                    } else {
                        return false;
                    }
                }
                if (c == '[' || c == '{') {
                    return true;
                } else {
                    return false;
                }
            }
        } catch (IOException ioex) {
        } finally {
            try {
                if (is != null) {
                    is.close();
                }
            } catch (IOException ioex1) {
            }
        }
        return false;
    }

    @Override
    public String getName() {
        return "JSON";
    }

    @Override
    public List<ValidationError> validate(StreamSource source, StreamSource schema) throws IOException {
        if (schema == null) {
            // Goal: perform syntax check with minimal memory consumption.
            // Solution: do not create objects and arrays while parsing.
            JSONParserLite parser = new JSONParserLite() {

                @Override
                public List newList() {
                    return null;
                }

                @Override
                public Map newObject() {
                    return null;
                }
            };
            try {
                parser.parse(source.getReader() != null ? source.getReader() : new InputStreamReader(source.getInputStream(), "UTF-8"));
            } catch (IOException ioex) {
                final String err = "" + ioex;
                return Collections.singletonList((ValidationError) new ValidationError(CRITICALITY.error) {

                    @Override
                    public String getMessage() {
                        return "JSON validation failed: " + err;
                    }
                });
            }
        } else {
            // load and validate using schema interpreted as JSON schema
            JSONParserLite parser = new JSONParserLite();
            try {
                // load source and schema
                Object data = parser.parse(source.getReader() != null ? source.getReader() : new InputStreamReader(source.getInputStream(), "UTF-8"));
                JSONSchema sch = new JSONSchema(schema.getReader() != null ? schema.getReader() : new InputStreamReader(schema.getInputStream(), "UTF-8"));
                // prepare validation context and validate
                JSONValidationContext vctx = new JSONValidationContext();
                boolean hasErrors = !sch.validate(vctx, data);
                // return found errors (if any)
                if (hasErrors) {
                    return vctx.getValidationErrors();
                }
            } catch (IOException ioex) {
                final String err = "" + ioex;
                return Collections.singletonList((ValidationError) new ValidationError(CRITICALITY.error) {

                    @Override
                    public String getMessage() {
                        return "JSON validation failed: " + err;
                    }
                });
            }
        }
        return Collections.EMPTY_LIST;
    }
}
