/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml.json.schema;

import com.ssg.tools.jsonxml.Formatter;
import com.ssg.tools.jsonxml.common.tools.BufferingReader;
import com.ssg.tools.jsonxml.json.JSONParserLite;
import com.ssg.tools.jsonxml.common.Path;
import java.io.File;
import java.io.InputStreamReader;
import java.io.StringWriter;
import java.net.URL;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.Date;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

/**
 * This class provides basic functionality on generation of data structures
 * based on JSON schema and generation rules.
 * 
 * @author ssg
 */
public class JSONDataGenerator {

    public static class GenerationContext extends JSONValidationContext {

        public static long MIN_TIME = 0;
        public static long MAX_TIME = 0;

        static {
            try {
                MIN_TIME = JSONValueFormat.F_TIME.parse("00:00:00").getTime();
                MAX_TIME = JSONValueFormat.F_TIME.parse("23:59:59").getTime();
            } catch (Throwable th) {
            }
        }

        /**
         * Creates schema element of specified type.
         * If type is "any" then randomly selects type.
         * 
         * @param schema
         * @param type
         * @param index
         * @return
         * @throws JSONSchemaException 
         */
        public Object create(JSchema schema, JSONValueType type, int index) throws JSONSchemaException {
            // System.out.println("create: " + type + " [" + index + "] " + schema);
            if (type == null) {
                return null;
            }

            if (schema.hasDefault() && index == 0) {
                return schema.getDefault();
            }
            if (type.equals(JSONValueType.STRING)) {
                return getRandomString(schema, index);
            } else if (type.equals(JSONValueType.NUMBER)) {
                return getRandomNumber(
                        schema.getMinimum(),
                        schema.getMaximum(),
                        schema.isExclusiveMinimum(),
                        schema.isExclusiveMaximum(),
                        false);
            } else if (type.equals(JSONValueType.INTEGER)) {
                return getRandomNumber(
                        schema.getMinimum(),
                        schema.getMaximum(),
                        schema.isExclusiveMinimum(),
                        schema.isExclusiveMaximum(),
                        true).intValue();
            } else if (type.equals(JSONValueType.BOOLEAN)) {
                return (Math.round(Math.random() * 25) % 2) == 0;
            } else if (type.equals(JSONValueType.ANY)) {
                int rt = (int) (Math.round(Math.random()) % (JSONValueType.PREDEFINED.length));
                if (rt < (JSONValueType.PREDEFINED.length - 1)) {
                    return create(schema, JSONValueType.PREDEFINED_TYPES[rt], index);
                } else {
                    return null;
                }
            } else if (type.equals(JSONValueType.ARRAY)) {
                return new ArrayList();
            } else if (type.equals(JSONValueType.OBJECT)) {
                return new LinkedHashMap();
            }
            return null;
        }

        /**
         * Returns default 3 of items.
         * Actual # of items is evaluated based on range in getItemscount().
         * 
         * @return 
         */
        public int getDefaultItemsCount() {
            return 25;
        }

        /**
         * Returns # of items based on default #, range, and random.
         * 
         * @param schema
         * @return
         * @throws JSONSchemaException 
         */
        public int getItemsCount(JSchema schema) throws JSONSchemaException {
            Integer min = schema.getMinItems();
            Integer max = schema.getMaxItems();

            boolean minUndefined = min == null;

            if (min == null) {
                min = 0;
            }
            if (max == null) {
                max = getDefaultItemsCount();
            }
            if (min > max) {
                if (minUndefined) {
                    min = max - getDefaultItemsCount();
                } else {
                    max = min + getDefaultItemsCount();
                }
            }

            int result = (int) (Math.round(Math.random() * max)) + min;
            if (result < min) {
                result = min;
            }
            if (result > max) {
                result = max;
            }
            return result;
        }

        /**
         * Returns true if item will be created.
         * Expect to return true for required and often randomly true for optional.
         * 
         * @param schema
         * @return
         * @throws JSONSchemaException 
         */
        public boolean willCreate(JSchema schema) throws JSONSchemaException {
            boolean result = schema.isRequired() || (Math.round(Math.random()) > 0.1) || getPath().getDepth() < 3;
            return result;
        }

        /**
         * evaluates item type for use with "create".
         * 
         * @param schema
         * @return
         * @throws JSONSchemaException 
         */
        public JSONValueType getItemType(JSchema schema) throws JSONSchemaException {
            List<JSONValueType> vt = schema.getType();
            if (vt == null) {
                return JSONValueType.ANY;
            }
            int rt = (int) (Math.round(Math.random()) % (vt.size()));
            if (rt < (vt.size())) {
                return vt.get(rt);
            } else {
                return vt.get(0);
            }
        }

        /**
         * Returns default range for numeric values. Used in getRandomNumber()
         * if none or partial limits are applied.
         * 
         * @return 
         */
        public double getDefaultRange() {
            return 1000;
        }

        /**
         * Returns number in specified range or, if range is not or partially specified,
         * in range evaluated from default range.
         * For excluding boundary values uses1 for integers, and 1/range for non-integers.
         * 
         * @param min
         * @param max
         * @param excludeMin
         * @param excludeMax
         * @param asInteger
         * @return 
         */
        public Number getRandomNumber(Number min, Number max, boolean excludeMin, boolean excludeMax, boolean asInteger) {
            if (min == null && max == null) {
                return (asInteger) ? (int) Math.round(Math.random() * getDefaultRange()) : Math.random() * getDefaultRange();
            } else {
                double d = Math.random();
                if (max == null) {
                    d = d * getDefaultRange() + min.doubleValue();
                    if (excludeMin && d <= min.doubleValue()) {
                        d = min.doubleValue() + ((asInteger) ? 1 : 1 / getDefaultRange());
                    }
                } else if (min == null) {
                    d = max.doubleValue() - d * getDefaultRange();
                    if (excludeMax && d >= max.doubleValue()) {
                        d = max.doubleValue() - ((asInteger) ? 1 : 1 / getDefaultRange());
                    }
                } else {
                    double range = max.doubleValue() - min.doubleValue();
                    d = d * range + min.doubleValue();
                    if (excludeMin) {
                        d += (asInteger) ? 1 : range / getDefaultRange();
                    }
                    if (excludeMax && d >= max.doubleValue()) {
                        d -= (asInteger) ? 1 : range / getDefaultRange();
                    }
                }
                return (asInteger) ? (int) Math.round(d) : d;
            }
        }

        /**
         * Should produce custom-formatted string for unknown formats.
         * 
         * @param schema
         * @param index
         * @return 
         */
        public Object getCustomString(JSchema schema, int index) {
            return null;
        }

        /**
         * Generates string based on limits and known formats. 
         * Uses "generator" to build schema-based format or getCustomString() when custom format detected.
         * 
         * @param schema
         * @param index
         * @return
         * @throws JSONSchemaException 
         */
        public Object getRandomString(JSchema schema, int index) throws JSONSchemaException {
            if (schema.getFormat() != null) {
                JSONValueFormat format = schema.getFormat();
                Integer min = schema.getMinLength();
                Integer max = schema.getMaxLength();
                if (min == null) {
                    min = 0;
                }
                if (max == null) {
                    max = 65535;
                }
                if (format.isSchema()) {
                    // schema - generate
                    return adjustString((new JSONDataGenerator()).generate(this, format.getSchema()),min,max);
                } else if (format.isCustom()) {
                    // custom format:... ?
                    return adjustString(getCustomString(schema, index),min,max);
                } else {
                    // predefined format
                    if (JSONValueFormat.DATE_TIME.equals(format)) {
                        long l = Math.round(Math.random() * (new Date()).getTime());
                        try {
                            return adjustString(JSONValueFormat.F_DATE_TIME.format(new Date(l)),min,max);
                        } catch (Throwable th) {
                        }
                    } else if (JSONValueFormat.DATE.equals(format)) {
                        long l = Math.round(Math.random() * (new Date()).getTime());
                        try {
                            return adjustString(JSONValueFormat.F_DATE.format(new Date(l)),min,max);
                        } catch (Throwable th) {
                        }
                    } else if (JSONValueFormat.TIME.equals(format)) {
                        long l = MIN_TIME + Math.round(Math.random() * (MAX_TIME - MIN_TIME));
                        try {
                            return adjustString(JSONValueFormat.F_TIME.format(new Date(l)),min,max);
                        } catch (Throwable th) {
                        }
                    } else if (JSONValueFormat.COLOR.equals(format)) {
                        return adjustString("" + JSONValueFormat.F_COLOR_NAME.values()[(int) Math.round(Math.random() * (JSONValueFormat.F_COLOR_NAME.values().length - 1))],min,max);
                    } else if (JSONValueFormat.STYLE.equals(format)) {
                        return adjustString("{ font-familty: Arial; }",min,max);
                    } else if (JSONValueFormat.EMAIL.equals(format)) {
                        return adjustString("aaa@bbb.ccc",min,max);
                    } else if (JSONValueFormat.PHONE.equals(format)) {
                        return adjustString("(111) 123 4567",min,max);
                    } else {
                        return adjustString(format.toString() + "_" + index,min,max);
                    }
                }

            } else {
                return "string_" + index;
            }
            return null;
        }

        public Object adjustString(Object object, int minLength, int maxLength) {
            if (object instanceof String) {
                String value = (String) object;

                if (value != null) {
                    if (value.length() > maxLength) {
                        value = value.substring(0, maxLength);
                    } else if (value.length() < minLength) {
                        StringBuilder sb = new StringBuilder(value);
                        for (int i = value.length(); i < minLength; i++) {
                            sb.append('_');
                        }
                        value = sb.toString();
                    }
                }
                return value;
            } else {
                return object;
            }
        }
    }

    public JSONDataGenerator() {
    }

    public Object generate(GenerationContext ctx, JSchema schema) throws JSONSchemaException {
        if (schema == null) {
            return null;
        }
        Path path = ctx.getPath();
        JSONValueType type = ctx.getItemType(schema);
        Object object = ctx.create(schema, type, 0);
        if (object instanceof Map) {
            Map map = (Map) object;
            if (schema.hasProperties()) {
                for (String key : schema.getProperties().keySet()) {
                    path.add(key);
                    JSchema subschema = schema.getProperty(key);
                    if (ctx.willCreate(subschema)) {
                        map.put(key, generate(ctx, subschema));
                    }
                    path.remove();
                }
            }
            return map;
        } else if (object instanceof List) {
            List list = (List) object;
            for (int i = 0; i < ctx.getItemsCount(schema); i++) {
                path.setIndex(i, true);
                //
                JSchema subschema = schema.getItems();
                Object item = generate(ctx, subschema);
                list.add(item);
            }
            path.setIndex(null, true);
            return list;
        } else {
            return object;
        }
    }

//    public static void main(String[] args) throws Exception {
//        //JSONSchema js = new JSONSchema(new URL("http://json-schema.org/card"));
//        JSONSchema js = new JSONSchema(new File("src/test/java/big.schema.json"));
//        if (1 == 0) {
//            String fileName = "./src/test/java/big.json";
//            Object sample = JSONParserLite.parse(new BufferingReader(new InputStreamReader((new File(fileName)).toURI().toURL().openStream(), "UTF-8")));
//            if (1 == 0) {
//                String url = "http://search.yahooapis.com/ImageSearchService/V1/imageSearch?appid=YahooDemo&query=Madonna&output=json";
//                url = "http://query.yahooapis.com/v1/public/yql?q=" + URLEncoder.encode("select * from upcoming.events", "UTF-8") + "&format=json";
//                url = "http://query.yahooapis.com/v1/public/yql?q=" + URLEncoder.encode("select * from music.artist.popular", "UTF-8") + "&format=json";
//                System.out.println("URL: " + url);
//                sample = JSONParserLite.parse(new BufferingReader(new InputStreamReader((new URL(url)).openStream(), "UTF-8")));
//            }
//            System.out.println("Sample data:\n" + Formatter.toJSONString(sample, true));
//            js = new JSONSchema();
//            JSONSchemaBuilder jsb = new JSONSchemaBuilder(js, "sample", sample, null);
//            StringWriter sw = new StringWriter();
//            js.save(sw, true);
//            System.out.println("Schema based on sample data:\n" + sw.toString());
//        }
//        JSONDataGenerator jg = new JSONDataGenerator();
//        Object o = jg.generate(new GenerationContext(), js.getSchema());
//        System.out.println("Generated sample data:\n" + Formatter.toJSONString(o, true));
//
//        JSONValidationContext vctx = new JSONValidationContext();
//        if (!js.validate(vctx, o)) {
//            System.out.println("Validation results for generated data:\n" + vctx.getSummary());
//        }
//
//    }
}
