/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml.json.schema;

import com.ssg.tools.jsonxml.common.ValidationContext;
import com.ssg.tools.jsonxml.common.ValidationException;
import com.ssg.tools.jsonxml.common.tools.URLResolver;
import com.ssg.tools.jsonxml.Formatter;
import com.ssg.tools.jsonxml.json.JSONParserLite;
import com.ssg.tools.jsonxml.json.JSONParserLiteContext;
import com.ssg.tools.jsonxml.json.JSONFormatterContext;
import com.ssg.tools.jsonxml.common.Formats;
import com.ssg.tools.jsonxml.common.ObjectsRegistry;
import com.ssg.tools.jsonxml.common.Path;
import com.ssg.tools.jsonxml.common.ValidationContext.ValidationSchema;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StringReader;
import java.io.Writer;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;
import javax.xml.transform.stream.StreamSource;

/**
 * JSON subSchema implementation.
 * Uses nested classes to represent and serialize/deserialize JSON subSchema.
 * - ATTRIBUTE - names of all attributes defined in subSchema. Names are prefixed with "_" if not starting with "$".
 * - JSONValueType - names of predefined types. Names are prefixed with "_" if not starting with "$".
 * - FORMATS - predefined formats. The "-" are replaced with "_".
 * - METHOD - predefined values for attribute "method".
 * - FRAGMENT_PROTOCOL - fragment protocol names. The "-" are replaced with "_".
 * - JBase - defines attribute values map, value setter, and presence checker.
 * - JSchema - JSON subSchema instance.
 * - JLink - extends JSchema to represent "links" element item (to separate properties no applicable to core schemas).
 * - JSONSchemaReflectiveBuilder - reflective builder extension to match Java and JSON representations on serialization/deserialization.
 * 
 * 
 * @author ssg
 */
public class JSONSchema implements ValidationSchema {

    public static String JSON_SCHEMA_URL = "http://json-schema.org/schema";
    private URL _url;
    private JSchema _schema;
    private Map<URI, JSchema> _dependencies = new HashMap<URI, JSchema>();
    private URLResolver _uriResolver = new URLResolver.CacheingURLResolver();

    public JSONSchema(URI uri, String encoding) throws JSONSchemaException {
        this(uri, encoding, null, null);
    }

    public JSONSchema(URI uri, String encoding, Map<URI, JSchema> dependencies, URLResolver uriResolver) throws JSONSchemaException {
        if (dependencies != null) {
            _dependencies = dependencies;
        }
        if (uriResolver != null) {
            _uriResolver = uriResolver;
        }
        try {
            _url = uri.toURL();
            _schema = load(_uriResolver.resolveURL(_url, encoding));
        } catch (MalformedURLException muex) {
            throw new JSONSchemaException("Failed to load JSON schema from " + uri + ": " + muex, muex);
        } catch (IOException ioex) {
            throw new JSONSchemaException("Failed to load JSON schema from " + _url + ": " + ioex, ioex);
        }
    }

    public JSONSchema(URI uri) throws JSONSchemaException {
        this(uri, "UTF-8");
    }

    public JSONSchema(URL url, String encoding) throws JSONSchemaException {
        _url = url;
        try {
            _schema = load(_uriResolver.resolveURL(_url, encoding));
        } catch (IOException ioex) {
            throw new JSONSchemaException("Failed to load JSON schema from " + _url + ": " + ioex, ioex);
        }
    }

    public JSONSchema(URL url) throws JSONSchemaException {
        this(url, "UTF-8");
    }

    public JSONSchema(File file, String encoding) throws JSONSchemaException {
        this(file.toURI(), encoding);
    }

    public JSONSchema(File file) throws JSONSchemaException {
        this(file.toURI());
    }

    public JSONSchema(String schema) throws JSONSchemaException {
        this(new StringReader(schema));
    }

    public JSONSchema(Reader reader) throws JSONSchemaException {
        _schema = load(new StreamSource(reader));
    }

    public JSONSchema(Reader reader, Map<URI, JSchema> dependencies) throws JSONSchemaException {
        if (dependencies != null) {
            _dependencies = dependencies;
        }
        _schema = load(new StreamSource(reader));
    }

    public JSONSchema(InputStream stream, String encoding) throws JSONSchemaException {
        try {
            _schema = load(new StreamSource(stream, encoding));
        } catch (IOException ioex) {
            throw new JSONSchemaException("Failed to load JSON schema from stream: " + ioex, ioex);
        }
    }

    public JSONSchema(InputStream stream) throws JSONSchemaException {
        this(stream, "UTF-8");
    }

    public JSONSchema(JSchema schema) throws JSONSchemaException {
        _schema = (schema != null) ? schema : new JSchema();
    }

    public JSONSchema() throws JSONSchemaException {
        _schema = new JSchema();
    }

    private Reader getStreamSourceReader(StreamSource source) throws JSONSchemaException {
        try {
            return (source.getReader() != null) ? source.getReader() : new InputStreamReader(source.getInputStream(), "UTF-8");
        } catch (Throwable th) {
            throw new JSONSchemaException("Failed to get reader from stream source. Error: " + th, th);
        }
    }

    public JSchema load(StreamSource source) throws JSONSchemaException {
        Reader r = getStreamSourceReader(source);
        JSONParserLite tjp = new JSONParserLite();
        JSONParserLiteContext ctx = new JSONParserLiteContext(r);
        try {
            tjp.parse(ctx);
        } catch (IOException ioex) {
            throw new JSONSchemaException("Failed to load JSON schema from " + ((_url != null) ? _url.toString() : " from string|reader") + ": " + ioex, ioex);
        }
        Object o = ctx.getObject();

        JSONSchemaReflectiveBuilder rb = new JSONSchemaReflectiveBuilder();
        Object ro = rb.reflectiveBuild(o, new Path("JSchema"), new Formats(), new ObjectsRegistry());
        if (ro instanceof JSchema) {
            if (((JSchema) ro).emptySchema()) {
                if (o instanceof Map) {
                    Map map = (Map) o;
                    if (map.containsKey("name")) {
                        if (map.size() > 1) {
                            throw new JSONSchemaException("Invalid JSON schema: " + ((_url != null) ? _url : o));
                        }
                    } else if (map.size() > 0) {
                        throw new JSONSchemaException("Invalid JSON schema: " + ((_url != null) ? _url : o));
                    }
                }
            }
        } else {
            throw new JSONSchemaException("Invalid schema: " + ((_url != null) ? _url : o));
        }
        loadRefs((JSchema) ro);
        return (JSchema) ro;
    }

    /**
     * Recursively scans for schema references ($ref with URLs) and loads them.
     * 
     * @param schema
     * @throws JSONSchemaException 
     */
    public void loadRefs(JSchema schema) throws JSONSchemaException {
        if (schema == null) {
            return;
        }
        if (schema.get$ref() != null) {
            // if schema starts with # - local path - nothing to load...
            if (schema.get$ref().toString().startsWith("#")) {
                return;
            }
            try {
                if (!_dependencies.containsKey(schema.get$ref())) {
                    //System.out.println("Loading schema for " + url);
                    JSONSchema js = new JSONSchema(schema.get$ref(), "UTF-8", _dependencies, _uriResolver);
                    _dependencies.put(schema.get$ref(), js.getSchema());
                    //System.out.println("Loaded  schema for " + url);
                }
            } catch (Throwable th) {
            }
        }
        Map<String, JSchema> ps = schema.getProperties();
        if (ps != null) {
            for (String key : ps.keySet()) {
                loadRefs(ps.get(key));
            }
        }
        ps = schema.getPatternProperties();
        if (ps != null) {
            for (String key : ps.keySet()) {
                loadRefs(ps.get(key));
            }
        }
        JSchema aps = schema.getAdditionalProperties();
        if (aps != null) {
            loadRefs(aps);
        }
    }

    public void save(Writer writer) throws IOException {
        save(writer, false);
    }

    public void save(Writer writer, boolean indented) throws IOException {
        JSONSchemaReflectiveBuilder rb = new JSONSchemaReflectiveBuilder();
        rb.mapClass("JSchema", JSchema.class);
        JSONFormatterContext format = new JSONFormatterContext(indented);
        format.setReflectiveBuilder(rb);
        Formatter.toFormattedString(writer, _schema, "", format, false);
    }

    /**
     * 
     * @param ctx
     * @param data
     * @return
     * @throws ValidationException 
     */
    public boolean validate(ValidationContext ctx, Object data) throws ValidationException {
        if (ctx instanceof JSONValidationContext) {
            return validate((JSONValidationContext) ctx, data);
        } else {
            throw new ValidationException("Need instance of JSONValidationContext, got " + ctx);
        }
    }

    /**
     * 
     * @param sample
     * @return 
     */
    public boolean validate(Object data) throws JSONSchemaException {
        return validate(new JSONValidationContext(_dependencies), getSchema(), data, true);
    }

    public boolean validate(JSONValidationContext ctx, Object data) throws JSONSchemaException {
        return validate(ctx, getSchema(), data, true);
    }

    /**
     * Validates data against provided subSchema. If recursive - validates each data property as well.
     * Accumulates errors in provided non-null list.
     * 
     * @param subSchema
     * @param data
     * @param recursive
     * @param errors - mandatory, must not be null
     * @return true if validation can continue or false if severe/critical error detected.
     */
    protected boolean validate(JSONValidationContext ctx, JSchema schema, Object data, boolean recursive) throws JSONSchemaException {
        if (schema == null) {
            return false;
        }

        return schema.validate(ctx, data, recursive);
    }

    public JSchema getSchema() {
        return _schema;
    }

    public void setSchema(JSchema schema) {
        _schema = schema;
    }

    /**
     * @return the _uriResolver
     */
    public URLResolver getUriResolver() {
        return _uriResolver;
    }

    /**
     * @param uriResolver the _uriResolver to set
     */
    public void setUriResolver(URLResolver uriResolver) {
        this._uriResolver = uriResolver;
    }
}
