/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml.json.schema;

import com.ssg.tools.jsonxml.common.Path;
import com.ssg.tools.jsonxml.common.ReflectiveBuilder;
import com.ssg.tools.jsonxml.common.StructureProcessingContext;
import com.ssg.tools.jsonxml.common.Utilities;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 *
 * @author ssg
 */
public class JSONSchemaBuilder {
    JSONSchema _schema;
    StructureProcessingContext _ctx;
    ReflectiveBuilder _rb;

    public JSONSchemaBuilder(JSONSchema schema, Object object) throws JSONSchemaException {
        this(schema, "schema", object, null);
    }

    public JSONSchemaBuilder(JSONSchema schema, String name, Object object) throws JSONSchemaException {
        this(schema, name, object, null);
    }

    public JSONSchemaBuilder(JSONSchema schema, String name, Object object, StructureProcessingContext ctx) throws JSONSchemaException {
        _ctx = (ctx != null) ? ctx : new StructureProcessingContext();
        _rb = _ctx.getReflectiveBuilder();
        _ctx.getPath().add(name);
        _schema = schema;
        JSchema js = createSchema();
        scan(js, _ctx.getPath(), object);
        if (js.getProperties() != null) {
            for (String key : js.getProperties().keySet()) {
                _schema.setSchema(js.getProperties().get(key));
                break;
            }
        }
        if (_schema != null && _schema.getSchema() != null) {
            _schema.getSchema().setName(name);
        }
    }

    public void scan(JSchema schema, Path path, Object object) throws JSONSchemaException {
        //System.out.println("scan: "+path);
        if (object == null) {
            return;
        }
        if (object instanceof List || object.getClass().isArray()) {
            JSchema obj = createSchema();
            obj.setType(JSONValueType.ARRAY);
            List<JSchema> items = new ArrayList<JSchema>();
            int idx = 0;
            for (Object item : (List) object) {
                JSchema iobj = createSchema();
                items.add(iobj);
                path.setIndex(idx++, true);
                scan(iobj, path, item);
            }
            path.setIndex(null, true);
            obj.setItems(itemsSchema(items));
            schema.setProperty(path.getLeafName(), obj);
        } else if (object instanceof Map) {
            JSchema obj = createSchema();
            obj.setType(JSONValueType.OBJECT);
            for (Object key : ((Map) object).keySet()) {
                path.add(key.toString());
                scan(obj, path, ((Map) object).get(key));
                path.remove();
            }
            schema.setProperty(path.getLeafName(), obj);
        } else if (Utilities.isScalarOrString(object)) {
            JSchema prop = createSchema();
            if (object instanceof String) {
                prop.setType(JSONValueType.STRING);
            } else if (object instanceof Boolean) {
                prop.setType(JSONValueType.BOOLEAN);
            } else if (Utilities.isNumericClass(object.getClass())) {
                Class nc = object.getClass();
                if (nc.equals(Double.class) || nc.equals(Float.class)) {
                    prop.setType(JSONValueType.NUMBER);
                } else {
                    prop.setType(JSONValueType.INTEGER);
                }
            }
            prop.setRequired(Boolean.TRUE);
            schema.setProperty(path.getLeafName(), prop);
        } else if (_rb != null && _rb.testIsReflectable(object)) {
            Map map = _rb.getReflectable(object);
            JSchema obj = createSchema();
            obj.setType(JSONValueType.OBJECT);
            for (Object key : map.keySet()) {
                path.add(key.toString());
                scan(obj, path, map.get(key));
                path.remove();
            }
            schema.setProperty(path.getLeafName(), obj);
        }
    }

    public JSchema itemsSchema(List<JSchema> items) throws JSONSchemaException {
        //System.out.println("itemsSchema: "+_ctx.getPath()+", size="+((items!=null) ? items.size() : 0));
        JSchema result = createSchema();
        for (JSchema b : items) {
            mergeSchemas(result, b, true);
        }
        if (result.getProperties() != null) {
            for (String key : result.getProperties().keySet()) {
                return result.getProperties().get(key);
            }
        }
        return null;
    }

    public void mergeSchemas(JSchema a, JSchema b, boolean restrictive) throws JSONSchemaException {
        // a == b ?
        for (ATTRIBUTE attr : a._attributes.keySet()) {
            if (b._attributes.containsKey(attr)) {
                Object va = a._attributes.get(attr);
                Object vb = b._attributes.get(attr);
                // merge
                switch (attr) {
                    case _type:
                        {
                            // join type sets
                            List<JSONValueType> all = new ArrayList<JSONValueType>();
                            if (va instanceof Iterable) {
                                for (Object vai : (Iterable) va) {
                                    if (!all.contains(vai)) {
                                        all.add((JSONValueType) vai);
                                    }
                                }
                            } else if (va != null) {
                                if (!all.contains(va)) {
                                    all.add((JSONValueType) va);
                                }
                            }
                            if (vb instanceof Iterable) {
                                for (Object vbi : (Iterable) vb) {
                                    if (!all.contains(vbi)) {
                                        all.add((JSONValueType) vbi);
                                    }
                                }
                            } else if (vb != null) {
                                if (!all.contains(vb)) {
                                    all.add((JSONValueType) vb);
                                }
                            }
                            // remove integer if number is present
                            if (all.contains(JSONValueType.NUMBER) && all.contains(JSONValueType.INTEGER)) {
                                all.remove(JSONValueType.INTEGER);
                            }
                            if (all.size() > 5 || all.size() > 1 && all.contains(JSONValueType.ANY)) {
                                all.clear();
                                all.add(JSONValueType.ANY);
                            }
                            a._attributes.put(attr, all);
                            break;
                        }
                    case _properties:
                    case _patternProperties:
                    case _additionalProperties:
                        {
                            Map<String, JSchema> all = (Map<String, JSchema>) va;
                            Map<String, JSchema> other = (Map<String, JSchema>) vb;
                            // merge matching
                            for (String key : all.keySet()) {
                                if (other.containsKey(key)) {
                                    mergeSchemas(all.get(key), other.get(key), restrictive);
                                } else {
                                    // ensure property is not mandatory
                                    JSchema v = all.get(key);
                                    if (v.isRequired()) {
                                        v.setRequired(false);
                                    }
                                }
                            }
                            // add missing as non-mandatory
                            for (String key : other.keySet()) {
                                if (!all.containsKey(key)) {
                                    JSchema v = other.get(key);
                                    if (v.isRequired()) {
                                        v.setRequired(false);
                                    }
                                    all.put(key, v);
                                }
                            }
                            break;
                        }
                    case _required:
                        {
                            if (a.isRequired() && !b.isRequired()) {
                                a.setRequired(false);
                            }
                            break;
                        }
                    case _items:
                        {
                            mergeSchemas((JSchema) va, (JSchema) vb, restrictive);
                            break;
                        }
                    case _additionalItems:
                    case _dependencies:
                    case _minimum:
                    case _maximum:
                    case _exclusiveMinimum:
                    case _exclusiveMaximum:
                    case _minItems:
                    case _maxItems:
                    case _uniqueItems:
                    case _pattern:
                    case _minLength:
                    case _maxLength:
                    case _enum:
                    case _default:
                    case _title:
                    case _description:
                    case _format:
                    case _divisibleBy:
                    case _disallow:
                    case _extends:
                    case _id:
                    case $ref:
                    case $schema:
                //////////////
                    case _links:
                    case _href:
                    case _rel:
                    case _targetSchema:
                    case _method:
                    case _enctype:
                    case _schema:
                    case _fragmentResolution:
                    case _readonly:
                    case _contentEncoding:
                    case _pathStart:
                    case _mediaType:
                    default:
                        throw new JSONSchemaException("Merge schemas error: unhandled attribute: " + attr);
                }
            } else {
                // TODO: decide if should be removed or modified - cusom method call?!
                a._attributes.remove(attr);
            }
        }
        // a <- b ?
        for (ATTRIBUTE attr : b._attributes.keySet()) {
            if (!a._attributes.containsKey(attr)) {
                // TODO: decide if should be added: custom method call?!
                a._attributes.put(attr, b._attributes.get(attr));
            }
        }
    }

    public JSONSchema getSchema() {
        return _schema;
    }

    public JSchema createSchema() {
        return new JSchema();
    }

    public JLink createLink() {
        return new JLink();
    }
    
}
