/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml.json.schema;

import com.ssg.tools.jsonxml.common.Formats;
import com.ssg.tools.jsonxml.common.ObjectsRegistry;
import com.ssg.tools.jsonxml.common.Path;
import com.ssg.tools.jsonxml.common.ReflectiveBuilder;
import com.ssg.tools.jsonxml.common.StructureProcessingException;
import com.ssg.tools.jsonxml.common.Utilities.Pair;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Adjusts subSchema class implementation specifics to JSON reflective parsing/formatting.
 * 
 */
/**
 *
 * @author ssg
 */
public class JSONSchemaReflectiveBuilder extends ReflectiveBuilder {

    /**
     * A_* constants for explicitly handled structures.
     */
    public static final String A_PROPERTIES = "Properties";
    public static final String A_DEPENDENCIES = "Dependencies";
    public static final String A_LINKS = "Links";
    public static final String A_EXTENDS = "Extends";
    public static final String A_TYPE = "Type";
    public static final String A_DISMISS = "Dismiss";
    public static final String A_FORMAT = "Format";

    public JSONSchemaReflectiveBuilder() {
        mapClass("JSchema", JSchema.class);
    }

    /**
     * Adjusts attribute names/values to Schema.
     * Detects and removes undefined values (e.g. those which have default values when not defined).
     * Converts internally defined names for enumerables into names in JSON subSchema specification.
     *
     * @param object
     * @return
     */
    @Override
    public Map<String, Object> getReflectable(Object object) {
        Map<String, Object> map = super.getReflectable(object);
        Map<String, Object> result = new LinkedHashMap<String, Object>();
        for (String key : map.keySet()) {
            Object val = map.get(key);
            if (object instanceof JSchema) {
                // remove undefined values:
                try {
                    String attrName = key.substring(0, 1).toLowerCase() + key.substring(1);
                    attrName = (!attrName.startsWith("$")) ? "_" + attrName : attrName;
                    ATTRIBUTE attr = ATTRIBUTE.valueOf(attrName);
                    if (!((JSchema) object).exists(attr)) {
                        val = null;
                    }
                } catch (Throwable th) {
                }
            }
            if (val == null) {
                continue;
            }
            if (val instanceof JSONMixedAttribute) {
                JSONMixedAttribute t = (JSONMixedAttribute) val;
                val = (t.isSchema()) ? t.getSchema() : t.getValue();
            } else if (val instanceof ATTRIBUTE) {
                String s = val.toString();
                if (s.startsWith("_")) {
                    s = s.substring(1);
                }
                val = s;
            } else if (val instanceof FRAGMENT_PROTOCOL) {
                val = ((FRAGMENT_PROTOCOL) val).name().replace("_", "-");
            }
            if (val instanceof Iterable) {
                boolean a_or_t = false;
                List l = new ArrayList();
                for (Object o : (Iterable) val) {
                    if (o != null) {
                        if (o instanceof ATTRIBUTE) {
                            a_or_t = true;
                            String s = o.toString();
                            if (s.startsWith("_")) {
                                s = s.substring(1);
                            }
                            o = s;
                        } else if (o instanceof JSONMixedAttribute) {
                            a_or_t = true;
                            if (((JSONMixedAttribute) o).isSchema()) {
                                o = ((JSONMixedAttribute) o).getSchema();
                            } else {
                                o = ((JSONMixedAttribute) o).getValue();
                            }
                        }
                    }
                    l.add(o);
                }
                if (l.size() == 1 && a_or_t) {
                    val = l.get(0);
                } else {
                    val = l;
                }
            }
            key = key.substring(0, 1).toLowerCase() + key.substring(1);
            result.put(key, val);
        }
        return result;
    }

    /**
     * Prepares string values for conversion into enums: adds underscore or replaces "-" with "_" when needed.
     *
     * @param enumClass
     * @param value
     * @return
     */
    @Override
    public Object string2EnumValue(Class enumClass, String value) {
        //System.out.println("string2EnumValue: c=" + enumClass + ", v=" + value);
        if (ATTRIBUTE.class.equals(enumClass) || JSONValueType.class.equals(enumClass)) {
            if (!value.startsWith("$") && !value.startsWith("_")) {
                value = '_' + value;
            }
        } else if (JSONValueFormat.class.equals(enumClass) || FRAGMENT_PROTOCOL.class.equals(enumClass)) {
            value = value.replace("-", "_");
        }
        //System.out.println("string2EnumValue: adjusted v=" + value);
        return super.string2EnumValue(enumClass, value);
    }

    /**
     * Provides proper initialization of nested structures for "unreliable" cases, i.e. collections with generics:
     * - for "properties" that is Map<String,JSchema>
     * - for "links" that is List<JLink>
     * - Tries to build URI when property type is URI.
     *
     * @param parent
     * @param object
     * @param clazz
     * @param path
     * @param formats
     * @param objectsRegistry
     * @return
     * @throws StructureProcessingException
     */
    @Override
    public Object reflectiveAdjust(Object parent, Object object, Class clazz, Path path, Formats formats, ObjectsRegistry objectsRegistry) throws StructureProcessingException {
        //System.out.println("reflectiveAdjust: for " + parent + " as " + clazz + " as " + path+" with "+object);
        String element = path.getLeafName();
        if (A_PROPERTIES.equals(element) && object instanceof Map) { // || element.equals("AdditionalProperties") || element.equals("PatternProperties")
            Map map = (Map) object;
            Map<String, JSchema> props = new LinkedHashMap<String, JSchema>();
            for (Object key : map.keySet()) {
                path.add((String) key);
                try {
                    Object val = reflectiveBuild(map.get(key), JSchema.class, path, formats, objectsRegistry);
                    props.put((String) key, (JSchema) val);
                } finally {
                    path.remove();
                }
            }
            object = props;
        } else if (A_DEPENDENCIES.equals(element) && object instanceof Map) {
            Map<String, Object> deps = new LinkedHashMap<String, Object>();
            Map map = (Map) object;
            for (Object key : map.keySet()) {
                if (!(key instanceof String)) {
                    throw new StructureProcessingException("Invalid dependency key: must be a string, found " + key.getClass() + " = " + key);
                }
                Object v = map.get(key);
                if (v instanceof String) {
                    deps.put((String) key, v);
                } else {
                    deps.put((String) key, reflectiveBuild(object, JSchema.class, path, formats, objectsRegistry));
                }
            }
            object = deps;
        } else if (A_LINKS.equals(element) && object instanceof List) {
            List list = new ArrayList();
            int idx = 0;
            for (Object o : (List) object) {
                path.setIndex(idx, true);
                try {
                    Object val = reflectiveBuild(o, JLink.class, path, formats, objectsRegistry);
                    list.add((JLink) val);
                } finally {
                }
            }
            path.setIndex(null, true);
            object = list;
        } else if (clazz.equals(URI.class) && object instanceof String) {
            try {
                object = new URI((String) object);
            } catch (URISyntaxException usex) {
                throw new StructureProcessingException("Failed to build URI for " + object + ": " + usex, usex);
            }
        } else if (A_EXTENDS.equals(element)) {
            object = this.createMixedAttribute(object, path, JSONValueExtends.class, formats, objectsRegistry);
        } else if (A_TYPE.equalsIgnoreCase(element) || A_DISMISS.equalsIgnoreCase(element)) {
            if (object instanceof List) {
                List list = new ArrayList();
                int idx = 0;
                for (Object o : (List) object) {
                    path.setIndex(idx++, true);
                    if (o instanceof String) {
                        list.add(new JSONValueType((String) o));
                    } else {
                        try {
                            JSchema val = (JSchema) reflectiveBuild(o, JSchema.class, path, formats, objectsRegistry);
                            list.add(new JSONValueType(val));
                        } catch (JSONSchemaException jsex) {
                            // TODO: ignore???
                        } finally {
                        }
                    }
                }
                path.setIndex(null, true);
                object = list;
            } else if (object instanceof String) {
                object = new JSONValueType((String) object);
            }
        } else if (A_FORMAT.equalsIgnoreCase(element)) {
            if (object instanceof List) {
                List list = new ArrayList();
                int idx = 0;
                for (Object o : (List) object) {
                    path.setIndex(idx, true);
                    if (o instanceof String) {
                        list.add(new JSONValueFormat((String) o));
                    } else {
                        try {
                            JSchema val = (JSchema) reflectiveBuild(o, JSchema.class, path, formats, objectsRegistry);
                            list.add(new JSONValueFormat(val));
                        } catch (JSONSchemaException jsex) {
                            // TODO: ignore???
                        } finally {
                        }
                    }
                }
                path.setIndex(null, true);
                object = list;
            } else if (object instanceof String) {
                object = new JSONValueFormat((String) object);
            }
        }
        return super.reflectiveAdjust(parent, object, clazz, path, formats, objectsRegistry);
    }

    /**
     * Correct readers for ambiguously defined attributes.
     * If attribute type is either boolean or JSchema, the reader for schema is preferred if not null.
     * 
     * If called with class (no object instance) filtering is not applied.
     * 
     * @param parent
     * @param propertyName
     * @param forWrite
     * @return 
     */
    @Override
    public Pair<Object, Class>[] getReflectableAccessors(Object parent, String propertyName, boolean forWrite) {
        // get default accessors set
        Pair<Object, Class>[] result = super.getReflectableAccessors(parent, propertyName, forWrite);
        // skip any processing if no data (parent is class)
        if (parent instanceof Class) {
            return result;
        }
        // instantiate cache of disabled accessors
        Set<Pair<Object, Class>> disabled = new HashSet<Pair<Object, Class>>();
        if (!forWrite) {
            int idxA = 0;
            for (Pair<Object, Class> pA : result) {
                if (disabled.contains(pA)) {
                    continue;
                }
                Method rA = (Method) pA.getA();
                if (pA.getA() instanceof Method) {
                    for (Pair<Object, Class> pB : result) {
                        if (disabled.contains(pB)) {
                            continue;
                        }
                        Method rB = (Method) pB.getA();
                        if (!rB.equals(rA) && (rB instanceof Method) && accessor2Key(rA).equals(accessor2Key(rB))) {
                            try {
                                if (Boolean.class.equals(pA.getB()) || boolean.class.equals(pA.getB())) {
                                    if (rB.invoke(parent) != null) {
                                        disabled.add(pA);
                                    } else {
                                        disabled.add(pB);
                                    }
                                    break;
                                } else if (Boolean.class.equals(pB.getB()) || Boolean.class.equals(pB.getB())) {
                                    if (rA.invoke(parent) != null) {
                                        disabled.add(pB);
                                    } else {
                                        disabled.add(pA);
                                    }
                                    break;
                                }
                            } catch (Throwable th) {
                                // we ignore this error though should never get it
                                //th.printStackTrace(System.out);
                            }
                        }
                    }
                }
                idxA++;
            }
        }
        // limit set of accessors if evaluated at least 1 disabled accessor
        if (disabled.size() > 0) {
            List<Pair<Object, Class>> l = new ArrayList<Pair<Object, Class>>(result.length - disabled.size());
            for (Pair<Object, Class> p : result) {
                if (!disabled.contains(p)) {
                    l.add(p);
                }
            }
            return (Pair<Object, Class>[]) l.toArray(new Pair[l.size()]);
        } else {
            return result;
        }
    }

    /**
     * Creation of mixed attribute.
     * Resolves problem of mixed attribute instantiation.
     * 
     * @param object
     * @param path
     * @param mixedClass
     * @param formats
     * @param objectsRegistry
     * @return
     * @throws StructureProcessingException 
     */
    public Object createMixedAttribute(Object object, Path path, Class mixedClass, Formats formats, ObjectsRegistry objectsRegistry) throws StructureProcessingException {
        if (object instanceof List) {
            List list = new ArrayList();
            int idx = 0;
            for (Object o : (List) object) {
                path.setIndex(idx, true);
                if (o instanceof String) {
                    list.add(createParameterizedObject(mixedClass, String.class, o));
                } else {
                    JSchema val = (JSchema) reflectiveBuild(o, JSchema.class, path, formats, objectsRegistry);
                    list.add(createParameterizedObject(mixedClass, JSchema.class, val));
                }
            }
            path.setIndex(null, true);
            object = list;
        } else if (object instanceof String) {
            object = createParameterizedObject(mixedClass, String.class, object);
        } else if (object != null) {
            object = reflectiveBuild(object, JSchema.class, path, formats, objectsRegistry);
            if (object instanceof JSchema) {
                object = createParameterizedObject(mixedClass, JSchema.class, object);
            }
        }
        return object;
    }

    /**
     * Creates object instance via constructor with parameter (mostly for mixed attributes support).
     * 
     * @param clazz
     * @param paramType
     * @param parameter
     * @return
     * @throws StructureProcessingException 
     */
    public static Object createParameterizedObject(Class clazz, Class paramType, Object parameter) throws StructureProcessingException {
        try {
            Class pc = (paramType != null) ? paramType : (parameter != null) ? parameter.getClass() : null;
            if (parameter instanceof JSchema) {
                pc = JSchema.class;
            }
            Constructor c = clazz.getConstructor(pc);
            return c.newInstance(parameter);
        } catch (Throwable th) {
            throw new StructureProcessingException("Failed to create parameterized instance for " + clazz + ", " + parameter);
        }
    }

    /**
     * Allows creation of non-empty mixed attribute objects.
     * 
     * @param object
     * @param clazz
     * @param path
     * @param values
     * @param formats
     * @param objectsRegistry
     * @return
     * @throws StructureProcessingException 
     */
    @Override
    public Object getObjectInstance(Object object, Class clazz, Path path, Map values, Formats formats, ObjectsRegistry objectsRegistry) throws StructureProcessingException {
        if (JSONMixedAttribute.class.isAssignableFrom(clazz)) {
            return this.reflectiveAdjust(object, values, clazz, path, formats, objectsRegistry);
        } else {
            return super.getObjectInstance(object, clazz, path, values, formats, objectsRegistry);
        }
    }
}
