/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml.json.schema;

import com.ssg.tools.jsonxml.common.CRITICALITY;
import com.ssg.tools.jsonxml.common.ValidationContext;
import com.ssg.tools.jsonxml.common.ValidationError;
import com.ssg.tools.jsonxml.common.ValidationException;
import com.ssg.tools.jsonxml.json.schema.JSONValidatorDefaultImpl.ListValidator;
import com.ssg.tools.jsonxml.json.schema.JSONValidatorDefaultImpl.ObjectValidator;
import com.ssg.tools.jsonxml.json.schema.JSONValidatorDefaultImpl.SchemaValidator;
import java.net.URI;
import java.util.EnumMap;
import java.util.LinkedHashMap;
import java.util.Locale;
import java.util.Map;

/**
 *
 * @author ssg
 */
public class JSONValidationContext extends ValidationContext {

    static Map<ATTRIBUTE, JSONAttributeValidator> _attributeValidators = new EnumMap<ATTRIBUTE, JSONAttributeValidator>(ATTRIBUTE.class);
    static JSONDataValidator _listValidator = new ListValidator();
    static JSONDataValidator _objectValidator = new ObjectValidator();
    static JSONDataValidator _schemaValidator = new SchemaValidator();
    boolean _assignMissingDefaults = false;
    boolean _enforceFormat = false;
    Map<URI, JSchema> _dependencies;

    public JSONValidationContext() {
    }

    public JSONValidationContext(Locale locale) {
        super(locale);
    }

    public JSONValidationContext(Locale locale, Map<URI, JSchema> dependencies) {
        super(locale);
        _dependencies = dependencies;
    }

    public JSONValidationContext(Map<URI, JSchema> dependencies) {
        _dependencies = dependencies;
    }

    public JSONAttributeValidator getValidator(ATTRIBUTE attr) {
        return _attributeValidators.get(attr);
    }

    public JSONDataValidator getValidator(Object data) {
        if (data instanceof Iterable) {
            return _listValidator;
        } else if (data instanceof Map || getReflectiveBuilder() != null && getReflectiveBuilder().testIsReflectable(data)) {
            return _objectValidator;
        } else {
            return null;
        }
    }

    public String getSummary() {
        StringBuilder sb = new StringBuilder();
        Map<ATTRIBUTE, Class> coverage = new LinkedHashMap<ATTRIBUTE, Class>();
        for (ATTRIBUTE attr : ATTRIBUTE.values()) {
            JSONDataValidator v = _attributeValidators.get(attr);
            if (v != null) {
                coverage.put(attr, v.getClass());
            } else {
                boolean ok = false;
                for (ATTRIBUTE oa : _objectValidator.getTestedAttributes()) {
                    if (attr.equals(oa)) {
                        coverage.put(attr, _objectValidator.getClass());
                        ok = true;
                        break;
                    }
                }
                if (!ok) {
                    for (ATTRIBUTE oa : _listValidator.getTestedAttributes()) {
                        if (attr.equals(oa)) {
                            coverage.put(attr, _listValidator.getClass());
                            ok = true;
                            break;
                        }
                    }
                }
                if (!ok) {
                    for (ATTRIBUTE oa : _schemaValidator.getTestedAttributes()) {
                        if (attr.equals(oa)) {
                            coverage.put(attr, _schemaValidator.getClass());
                            ok = true;
                            break;
                        }
                    }
                }
                if (!ok) {
                    for (JSONDataValidator va : _attributeValidators.values()) {
                        for (ATTRIBUTE oa : va.getTestedAttributes()) {
                            if (attr.equals(oa)) {
                                coverage.put(attr, va.getClass());
                                ok = true;
                                break;
                            }
                        }
                        if (ok) {
                            break;
                        }
                    }
                }
                if (!ok) {
                    coverage.put(attr, null);
                }
            }
        }
        int covered = 0;
        for (Object o : coverage.keySet()) {
            if (coverage.get(o) != null) {
                covered++;
            }
        }
        sb.append("JSON schema attributes validation coverage (" + covered + "/" + ATTRIBUTE.values().length + "):\n");
        for (ATTRIBUTE attr : coverage.keySet()) {
            String s = attr.name();
            while (s.length() < 30) {
                s += " ";
            }
            Class c = coverage.get(attr);
            sb.append("  " + s + ": " + ((c != null) ? c.getName() : "-- no validator --") + "\n");
        }
        int eCount = 0;
        int wCount = 0;
        for (ValidationError err : getValidationErrors()) {
            if (err.getCriticality().equals(CRITICALITY.warning)) {
                wCount++;
            } else {
                eCount++;
            }
        }

        sb.append("Errors summary E/W/A [" + eCount + "/" + wCount + "/" + getValidationErrors().size() + "]:\n");
        for (ValidationError err : getValidationErrors()) {
            sb.append("  " + err.getMessage() + "\n");
        }

        return sb.toString();
    }

    /**
     * Initialize default attribute validators from JSONValidatorDefaultImpl class.
     */
    static {
        try {
            // find and enforce registration of all AttributeValidator implementations...
            for (Class c : JSONValidatorDefaultImpl.class.getDeclaredClasses()) {
                if (JSONAttributeValidator.class.isAssignableFrom(c)) {
                    JSONAttributeValidator v = (JSONAttributeValidator) c.newInstance();
                    _attributeValidators.put(v.getAttribute(), v);
                }
            }
        } catch (Throwable th) {
        }
    }

    /**
     * @return the _assignMissingDefaults
     */
    public boolean isAssignMissingDefaults() {
        return _assignMissingDefaults;
    }

    /**
     * @param assignMissingDefaults the _assignMissingDefaults to set
     */
    public void setAssignMissingDefaults(boolean assignMissingDefaults) {
        this._assignMissingDefaults = assignMissingDefaults;
    }

    /**
     * @return the _enforceFormat
     */
    public boolean isEnforceFormat() {
        return _enforceFormat;
    }

    /**
     * @param enforceFormat the _enforceFormat to set
     */
    public void setEnforceFormat(boolean enforceFormat) {
        this._enforceFormat = enforceFormat;
    }

    /**
     * @return the _dependencies
     */
    public Map<URI, JSchema> getDependencies() {
        return _dependencies;
    }

    /**
     * @param dependencies the _dependencies to set
     */
    public void setDependencies(Map<URI, JSchema> dependencies) {
        this._dependencies = dependencies;
    }

    public JSchema getDependency(JSchema schema) {
        try {
            if (schema == null || _dependencies == null || schema.get$ref() == null) {
                return null;
            }
            return _dependencies.get(schema.get$ref());
        } catch (JSONSchemaException jsex) {
            return null;
        }
    }

    @Override
    public ValidationSchema getValidationSchema(URI uri, String encoding) throws ValidationException {
        JSONSchema schema = new JSONSchema();
        if (getUrlResolver() != null) {
            schema.setUriResolver(this.getUrlResolver());
        }
        schema.setSchema(schema.load(schema.getUriResolver().resolveURI(uri, encoding)));
        return schema;
    }

    @Override
    public boolean validate(ValidationSchema schema, Object data) throws ValidationException {
        if (schema == null) {
            throw new ValidationException("No JSON schema provided for validation.");
        }
        return schema.validate(this, data);
    }
}
