/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml.json.schema;

import com.ssg.tools.jsonxml.common.Path;
import com.ssg.tools.jsonxml.common.Utilities;
import com.ssg.tools.jsonxml.json.schema.JSONValidatorDefaultImpl.AV_Type;
import java.net.URI;
import java.net.URISyntaxException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

/**
 * This class provides set of nested validator implementations
 * that are used by default in ValidationContext and automatically registered by it.
 * 
 * also default Object and List validators are implemented here.
 * 
 * @author ssg
 */
public class JSONValidatorDefaultImpl {

    /**
     * Compares objects as specified in JSON schema for uniqueItems detection.
     * 
     * @param ctx
     * @param a
     * @param b
     * @param depth
     * @return
     * @throws JSONSchemaException 
     */
    public static boolean same(JSONValidationContext ctx, Object a, Object b, int depth) throws JSONSchemaException {
        if (depth > 100) {
            throw new JSONSchemaException("Can't compare object to depth deeper than 100 or self-reference(s) exist.");
        }
        // nulls rule
        if (a == null && b == null) {
            return true;
        }
        if (a == null || b == null) {
            return false;
        }
        // scalars rule
        if (Utilities.isScalarOrString(a)) {
            if (a.equals(b)) {
                return true;
            } else {
                return false;
            }
        }
        // lists rule
        if (a instanceof Iterable && b instanceof Iterable) {
            Iterator ia = ((Iterable) a).iterator();
            Iterator ib = ((Iterable) b).iterator();
            while (ia.hasNext()) {
                if (!ib.hasNext()) {
                    return false;
                }
                Object oa = ia.next();
                Object ob = ib.next();
                if (!same(ctx, oa, ob, depth + 1)) {
                    return false;
                }
            }
            if (ib.hasNext()) {
                return false;
            }
            return true;
        }
        // properties rule
        Map ma = ctx.toMap(a);
        Map mb = ctx.toMap(b);
        if (ma != null && mb != null) {
            if (ma.size() != mb.size()) {
                return false;
            }
            for (Object key : ma.keySet()) {
                Object oa = ma.get(key);
                Object ob = mb.get(key);
                if (!same(ctx, oa, ob, depth + 1)) {
                    return false;
                }
            }
            return true;
        } else {
            return false;
        }
    }

    public static class AV_Type extends JSONAttributeValidator {

        boolean _disallow = false;

        public ATTRIBUTE getAttribute() {
            return ATTRIBUTE._type;
        }

        public boolean validate(JSONValidationContext ctx, Object data, JSchema schema) throws JSONSchemaException {
            if (schema == null) {
                throw new JSONSchemaException(getClass().getName() + "Invalid validator call: schema is null.");
            }
            // we do not check untyped data. For null required is responsible!
            if (data == null) {
                return true;
            }
            boolean ok = false;
            Class clazz = data.getClass();
            List<JSONValueType> types = (_disallow) ? schema.getDisallow() : schema.getType();
            // default type is any - check it here.
            if (types == null || types.isEmpty() || types.contains(JSONValueType.ANY)) {
                return true;
            }
            for (JSONValueType type : types) {
                if (ok) {
                    break;
                }
                if (type.equals(JSONValueType.ANY)) {
                    ok = true;
                } else if (type.equals(JSONValueType.STRING)) {
                    if (data instanceof String) {
                        ok = true;
                    }
                } else if (type.equals(JSONValueType.NUMBER)) {
                    if (data instanceof Number) {
                        ok = true;
                    }
                } else if (type.equals(JSONValueType.INTEGER)) {
                    if (data instanceof Number) {
                        ok = data instanceof Byte
                                || data instanceof Short
                                || data instanceof Integer
                                || data instanceof Long;
                    }
                } else if (type.equals(JSONValueType.BOOLEAN)) {
                    if (data instanceof Boolean) {
                        ok = true;
                    }
                } else if (type.equals(JSONValueType.ARRAY)) {
                    if (data instanceof List || clazz.isArray()) {
                        ok = true;
                    }
                } else if (type.equals(JSONValueType.OBJECT)) {
                    if (data instanceof Map) {
                        ok = true;
                    } else if (!Utilities.isScalarOrString(data) && !(data instanceof List) && !clazz.isArray()) {
                        ok = true;
                    }
                } else if (type.equals(JSONValueType.NULL)) {
                    // unreachable... nulls are not checked here...
                } else if (type.isSchema()) {
                    // validate using schema...
                    if (type.getSchema().validate(ctx, data, true)) {
                        ok = true;
                    }
                }
            }

            if (_disallow) {
                ok = !ok;
            }
            if (!ok) {
                ctx.addValidationError(new JSONValidationError(ctx.getPath(), data, schema, getAttribute(), "Invalid data type (" + ((data != null) ? data.getClass().getName() : "null") + "). " + ((_disallow) ? "Disallowed " : "Expected ") + types + "."));
            }
            return ok;
        }
    }

    public static class AV_Required extends JSONAttributeValidator {

        public boolean validate(JSONValidationContext ctx, Object data, JSchema schema) throws JSONSchemaException {
            if (schema.isRequired() && data == null) {
                ctx.addValidationError(new JSONValidationError(ctx.getPath(), data, schema, getAttribute(), "Missing mandatory value."));
                return false;
            }
            return true;
        }

        public ATTRIBUTE getAttribute() {
            return ATTRIBUTE._required;
        }
    }

    public static class AV_Dependencies extends JSONAttributeValidator {

        @Override
        public ATTRIBUTE getAttribute() {
            return ATTRIBUTE._dependencies;
        }

        @Override
        public boolean validate(JSONValidationContext ctx, Object data, JSchema schema) throws JSONSchemaException {
            if (ctx == null || schema == null) {
                throw new JSONSchemaException("Invalid context or schema: neither can be null.");
            }
            Map<String, Object> deps = schema.getDependencies();
            if (deps == null || deps.isEmpty() || data == null) {
                return true;
            }
            Map map = ctx.toMap(data);
            if (map == null) {
                ctx.addValidationError(new JSONValidationError(ctx.getPath(), data, schema, getAttribute(), "Failed to test dependencies for non-object."));
                return false;
            }
            boolean ok = true;
            for (String dep : deps.keySet()) {
                // TODO: implement dependency check for String and JSchema values
                if (map.containsKey(dep)) {
                    Object v = deps.get(dep);
                    if (v instanceof JSchema) {
                        if (!((JSchema) v).validate(ctx, map.get(dep), true)) {
                            ctx.addValidationError(new JSONValidationError(ctx.getPath(), data, schema, getAttribute(), "Dependency validatione failed: " + dep));
                            ok = false;
                        }
                    } else if (v instanceof String) {
                        if (!map.containsKey(v)) {
                            ctx.addValidationError(new JSONValidationError(ctx.getPath(), data, schema, getAttribute(), "Missing dependency: " + dep));
                            ok = false;
                        }
                    } else {
                        throw new JSONSchemaException("Invalid dependency value: must be schema or string, found " + v);
                    }
                }
            }
            return ok;
        }
    }

    public static class AV_Minimum extends JSONAttributeValidator {

        @Override
        public ATTRIBUTE getAttribute() {
            return ATTRIBUTE._minimum;
        }

        @Override
        public boolean validate(JSONValidationContext ctx, Object data, JSchema schema) throws JSONSchemaException {
            if (ctx == null || schema == null) {
                throw new JSONSchemaException("Invalid context or schema: neither can be null.");
            }
            // we ignore limits for undefined data.
            if (data == null || !schema.exists(getAttribute())) {
                return true;
            }
            Number min = schema.getMinimum();
            boolean exclusive = schema.isExclusiveMinimum();
            if (Utilities.isNumericClass(data.getClass())) {
                double val = ((Number) data).doubleValue();
                double mind = min.doubleValue();
                int res = (val < mind) ? -1 : (val == mind) ? 0 : 1;
                if (res < 0) {
                    ctx.addValidationError(new JSONValidationError(ctx.getPath(), data, schema, getAttribute(), "Value is less than minimum [" + min + "]: " + data + "."));
                    return false;
                } else if (exclusive && res == 0) {
                    ctx.addValidationError(new JSONValidationError(ctx.getPath(), data, schema, getAttribute(), "Value equals to exclusive minimum [" + min + "]."));
                    return false;
                }
            } else {
                //ctx.addValidationError(new ValidationError(ctx.getPath(),data,schema,getAttribute(),"Cannot check minimum for not a number."));
                // Not applicable for non-numeric values
            }
            return true;
        }

        @Override
        public Iterable<ATTRIBUTE> getTestedAttributes() {
            List<ATTRIBUTE> l = new ArrayList<ATTRIBUTE>(2);
            l.add(getAttribute());
            l.add(ATTRIBUTE._exclusiveMinimum);
            return l;
        }
    }

    public static class AV_Maximum extends JSONAttributeValidator {

        @Override
        public ATTRIBUTE getAttribute() {
            return ATTRIBUTE._maximum;
        }

        @Override
        public boolean validate(JSONValidationContext ctx, Object data, JSchema schema) throws JSONSchemaException {
            if (ctx == null || schema == null) {
                throw new JSONSchemaException("Invalid context or schema: neither can be null.");
            }
            // we ignore limits for undefined data.
            if (data == null || !schema.exists(getAttribute())) {
                return true;
            }
            Number max = schema.getMaximum();
            boolean exclusive = schema.isExclusiveMaximum();
            if (Utilities.isNumericClass(data.getClass())) {
                double val = ((Number) data).doubleValue();
                double maxd = max.doubleValue();
                int res = (val > maxd) ? 1 : (val == maxd) ? 0 : -1;
                if (res > 0) {
                    ctx.addValidationError(new JSONValidationError(ctx.getPath(), data, schema, getAttribute(), "Value is greater than maximum [" + max + "]: " + data + "."));
                    return false;
                } else if (exclusive && res == 0) {
                    ctx.addValidationError(new JSONValidationError(ctx.getPath(), data, schema, getAttribute(), "Value equals to exclusive maximum [" + max + "]."));
                    return false;
                }
            } else {
                //ctx.addValidationError(new ValidationError(ctx.getPath(),data,schema,getAttribute(),"Cannot check maximum for not a number."));
                // Not applicable for non-numeric values
            }
            return true;
        }

        @Override
        public Iterable<ATTRIBUTE> getTestedAttributes() {
            List<ATTRIBUTE> l = new ArrayList<ATTRIBUTE>(2);
            l.add(getAttribute());
            l.add(ATTRIBUTE._exclusiveMaximum);
            return l;
        }
    }

    public static class AV_MinItems extends JSONAttributeValidator {

        @Override
        public ATTRIBUTE getAttribute() {
            return ATTRIBUTE._minItems;
        }

        @Override
        public boolean validate(JSONValidationContext ctx, Object data, JSchema schema) throws JSONSchemaException {
            if (ctx == null || schema == null) {
                throw new JSONSchemaException("Invalid context or schema: neither can be null.");
            }
            Collection coll = ctx.toCollection(data);
            if (coll == null || !schema.exists(getAttribute())) {
                return true;
            }

            Number min = schema.getMinItems();
            Integer count = coll.size();
            if (count < min.intValue()) {
                ctx.addValidationError(new JSONValidationError(ctx.getPath(), data, schema, getAttribute(), "Too few elements in array. Minimum is " + min + ", actually " + count + "."));
                return false;
            } else {
                return true;
            }
        }
    }

    public static class AV_MaxItems extends JSONAttributeValidator {

        @Override
        public ATTRIBUTE getAttribute() {
            return ATTRIBUTE._maxItems;
        }

        @Override
        public boolean validate(JSONValidationContext ctx, Object data, JSchema schema) throws JSONSchemaException {
            if (ctx == null || schema == null) {
                throw new JSONSchemaException("Invalid context or schema: neither can be null.");
            }
            Collection coll = ctx.toCollection(data);
            if (coll == null || !schema.exists(getAttribute())) {
                return true;
            }

            Number max = schema.getMaxItems();
            Integer count = coll.size();
            if (count > max.intValue()) {
                ctx.addValidationError(new JSONValidationError(ctx.getPath(), data, schema, getAttribute(), "Too many elements in array. Maximum is " + max + ", actually " + count + "."));
                return false;
            } else {
                return true;
            }
        }
    }

    /**
     * Implements logic for uniqueItems attribute.
     * Supports 2 modes: stop validation on 1st error or find all duplicates.
     */
    public static class AV_UniqueItems extends JSONAttributeValidator {

        private boolean _stopOnFirstError = true;

        @Override
        public ATTRIBUTE getAttribute() {
            return ATTRIBUTE._uniqueItems;
        }

        @Override
        public boolean validate(JSONValidationContext ctx, Object data, JSchema schema) throws JSONSchemaException {
            if (ctx == null || schema == null) {
                throw new JSONSchemaException("Invalid context or schema: neither can be null.");
            }
            Collection coll=ctx.toCollection(data);
            if (coll == null || !schema.exists(getAttribute())) {
                return true;
            }
            //System.out.println("validate "+getAttribute()+" for "+data);
            if (!schema.isUniqueItems()) {
                return true;
            }

            boolean ok = true;
            boolean allDuplicates = !isStopOnFirstError();
            Set<String> duplicateIdxs = (allDuplicates) ? new HashSet<String>() : null;

            Iterator ia = coll.iterator();
            int idxA = 0;
            while (ia.hasNext()) {
                Object oa = ia.next();

                Iterator ib = coll.iterator();
                int idxB = 0;
                while (ib.hasNext()) {
                    Object ob = ib.next();
                    if (idxA != idxB) {
                        if (same(ctx, oa, ob, 0)) {
                            boolean createError = true;
                            String key = null;
                            if (allDuplicates) {
                                key = Math.min(idxA, idxB) + "-" + Math.max(idxA, idxB);
                                if (duplicateIdxs.contains(key)) {
                                    createError = false;
                                } else {
                                    duplicateIdxs.add(key);
                                }
                            }
                            if (createError) {
                                ctx.addValidationError(new JSONValidationError(ctx.getPath(), data, schema, getAttribute(), "Duplicate items in array at [" + idxA + "] and [" + idxB + "]."));
                            }
                            ok = false;
                            if (isStopOnFirstError()) {
                                return ok;
                            }
                        }
                    }
                    idxB++;
                }
                idxA++;
            }
            return ok;
        }

        /**
         * @return the _stopOnFirstError
         */
        public boolean isStopOnFirstError() {
            return _stopOnFirstError;
        }

        /**
         * @param stopOnFirstError the _stopOnFirstError to set
         */
        public void setStopOnFirstError(boolean stopOnFirstError) {
            this._stopOnFirstError = stopOnFirstError;
        }
    }

    public static class AV_Pattern extends JSONAttributeValidator {

        @Override
        public ATTRIBUTE getAttribute() {
            return ATTRIBUTE._pattern;
        }

        @Override
        public boolean validate(JSONValidationContext ctx, Object data, JSchema schema) throws JSONSchemaException {
            if (ctx == null || schema == null) {
                throw new JSONSchemaException("Invalid context or schema: neither can be null.");
            }
            // we ignore patterns for non-string data.
            if (!(data instanceof String) || !schema.exists(getAttribute())) {
                return true;
            }

            if (!((String) data).matches(schema.getPattern())) {
                ctx.addValidationError(new JSONValidationError(ctx.getPath(), data, schema, getAttribute(), "Value does not match regular expression [" + schema.getPattern() + "]: " + data));
                return false;
            } else {
                return true;
            }
        }
    }

    public static class AV_MinLength extends JSONAttributeValidator {

        @Override
        public ATTRIBUTE getAttribute() {
            return ATTRIBUTE._minLength;
        }

        @Override
        public boolean validate(JSONValidationContext ctx, Object data, JSchema schema) throws JSONSchemaException {
            if (ctx == null || schema == null) {
                throw new JSONSchemaException("Invalid context or schema: neither can be null.");
            }
            // we ignore min length for non-string data.
            if (!(data instanceof String) || !schema.exists(getAttribute())) {
                return true;
            }

            Integer min = schema.getMinLength();
            if (((String) data).length() < min) {
                ctx.addValidationError(new JSONValidationError(ctx.getPath(), data, schema, getAttribute(), "Too short string [" + min + "]: " + ((String) data).length() + ", " + data + "."));
                return false;
            } else {
                return true;
            }
        }
    }

    public static class AV_MaxLength extends JSONAttributeValidator {

        @Override
        public ATTRIBUTE getAttribute() {
            return ATTRIBUTE._maxLength;
        }

        @Override
        public boolean validate(JSONValidationContext ctx, Object data, JSchema schema) throws JSONSchemaException {
            if (ctx == null || schema == null) {
                throw new JSONSchemaException("Invalid context or schema: neither can be null.");
            }
            // we ignore min length for non-string data.
            if (!(data instanceof String) || !schema.exists(getAttribute())) {
                return true;
            }

            Integer max = schema.getMaxLength();
            if (((String) data).length() > max) {
                ctx.addValidationError(new JSONValidationError(ctx.getPath(), data, schema, getAttribute(), "Too long string [" + max + "]: " + ((String) data).length() + ", " + data + "."));
                return false;
            } else {
                return true;
            }
        }
    }

    public static class AV_Enum extends JSONAttributeValidator {

        @Override
        public ATTRIBUTE getAttribute() {
            return ATTRIBUTE._enum;
        }

        @Override
        public boolean validate(JSONValidationContext ctx, Object data, JSchema schema) throws JSONSchemaException {
            if (ctx == null || schema == null) {
                throw new JSONSchemaException("Invalid context or schema: neither can be null.");
            }
            // we ignore min length for non-string data.
            if (!schema.exists(getAttribute())) {
                return true;
            }

            List values = schema.getEnum();
            if (!values.contains(data)) {
                if (data instanceof String && values.get(0).getClass().isEnum()) {
                    for (Object o : values) {
                        if (data.equals(o.toString())) {
                            return true;
                        }
                    }
                }
                ctx.addValidationError(new JSONValidationError(ctx.getPath(), data, schema, getAttribute(), "Value is not in list [" + values + "]: " + data + "."));
                return false;
            } else {
                return true;
            }
        }
    }

    public static class AV_Disallow extends AV_Type {

        @Override
        public ATTRIBUTE getAttribute() {
            return ATTRIBUTE._disallow;
        }

        public AV_Disallow() {
            _disallow = true;
        }
    }

    public static class AV_DivisibleBy extends JSONAttributeValidator {

        @Override
        public ATTRIBUTE getAttribute() {
            return ATTRIBUTE._divisibleBy;
        }

        @Override
        public boolean validate(JSONValidationContext ctx, Object data, JSchema schema) throws JSONSchemaException {
            if (ctx == null || schema == null) {
                throw new JSONSchemaException("Invalid context or schema: neither can be null.");
            }
            // we ignore min length for non-string data.
            if (!(data instanceof Number) || !schema.exists(getAttribute())) {
                return true;
            }

            Number divider = schema.getDivisibleBy();
            if (divider.equals(0)) {
                throw new JSONSchemaException("Divider SHOULD NOT be 0 in schema " + schema);
            }

            double result = ((Number) data).doubleValue() / divider.doubleValue();
            if ((result - (long) result) != 0) {
                ctx.addValidationError(new JSONValidationError(ctx.getPath(), data, schema, getAttribute(), "Value is not divisible by [" + divider + "] without reminder: " + data + ", reminder is " + (result - (long) result) + "."));
                return false;
            } else {
                return true;
            }
        }
    }

    public static class AV_Format extends JSONAttributeValidator {

        @Override
        public ATTRIBUTE getAttribute() {
            return ATTRIBUTE._format;
        }

        @Override
        public boolean validate(JSONValidationContext ctx, Object data, JSchema schema) throws JSONSchemaException {
            if (ctx == null || schema == null) {
                throw new JSONSchemaException("Invalid context or schema: neither can be null.");
            }
            // we ignore nulls
            if (data == null || !schema.exists(getAttribute())) {
                return true;
            }

            JSONValueFormat fmt = schema.getFormat();

            try {
                if (fmt.equals(JSONValueFormat.DATE_TIME)) {
                    if (data instanceof String) {
                        Date date = JSONValueFormat.F_DATE_TIME.parse((String) data);
                        return true;
                    } else if (data instanceof Date) {
                        return true;
                    }
                } else if (fmt.equals(JSONValueFormat.DATE)) {
                    if (data instanceof String) {
                        Date date = JSONValueFormat.F_DATE.parse((String) data);
                        return true;
                    } else if (data instanceof Date) {
                        return true;
                    }
                } else if (fmt.equals(JSONValueFormat.TIME)) {
                    if (data instanceof String) {
                        Date date = JSONValueFormat.F_TIME.parse((String) data);
                        return true;
                    } else if (data instanceof Date) {
                        return true;
                    }
                } else if (fmt.equals(JSONValueFormat.UTC_MILLISEC)) {
                    if (data instanceof Number && ((Number) data).longValue() >= 0) {
                        return true;
                    }
                } else if (fmt.equals(JSONValueFormat.REGEX)) {
                    if (data instanceof String) {
                        Pattern p = Pattern.compile((String) data);
                        return true;
                    }
                } else if (fmt.equals(JSONValueFormat.COLOR)) {
                    if (data instanceof JSONValueFormat.F_COLOR_NAME) {
                        return true;
                    }
                    // http://www.w3.org/TR/2007/CR-CSS21-20070719
                    // TODO: adjust regular expressions for each variant
                    if (data instanceof String && ((String) data).length() > 1) {
                        String color = (String) data;
                        if (color.startsWith("#")) {
                            // #rgb or #rrggbb
                            // TODO: validate properly #rrggbb or #rgb. Now - draftly...
                            if (color.matches("#[0-9,a-f,A-F]{3}") || color.matches("#[0-9,a-f,A-F]{6}")) {
                                return true;
                            }
                        } else if (color.startsWith("rgb(")) {
                            // rgb(r,g,b)
                            // TODO: validate properly rgb(r,g,b). Now - draftly...
                            if (color.matches("rgb\\(\\b(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\,){2}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\b\\)")) {
                                return true;
                            }
                        } else if (JSONValueFormat.F_COLOR_NAME.valueOf(color) != null) {
                            // predefined color
                            return true;
                        }
                    }
                } else if (fmt.equals(JSONValueFormat.STYLE)) {
                    if (data instanceof String) {
                        return ((String) data).startsWith("{") && ((String) data).endsWith("}");
                    }
                } else if (fmt.equals(JSONValueFormat.PHONE)) {
                    // check for E.123 format
                    //System.out.println(s+": "+s.matches("[0-9][ ,0-9]*( ext. [0-9]+)*")+", "+s.matches("\\([0-9]+\\)[ ,\\-,0-9]+( ext. [0-9]+)*")+", "+s.matches("\\+[0-9]+[ ,0-9]+( ext. [0-9]+)*"));
                    //System.out.println(s+": "+s.matches("[0-9][ ,0-9]*( ext. [0-9]+)*|\\([0-9]+\\)[ ,\\-,0-9]+( ext. [0-9]+)*|\\+[0-9]+[ ,0-9]+( ext. [0-9]+)*"));
                    if (data instanceof String && ((String) data).matches("[0-9][ ,0-9]*( ext. [0-9]+)*|\\([0-9]+\\)[ ,\\-,0-9]+( ext. [0-9]+)*|\\+[0-9]+[ ,0-9]+( ext. [0-9]+)*")
                            || data instanceof Number) {
                        return true;
                    }
                } else if (fmt.equals(JSONValueFormat.URI)) {
                    // check for URI
                    try {
                        new URI((String) data);
                        return true;
                    } catch (URISyntaxException usex) {
                    }
                } else if (fmt.equals(JSONValueFormat.EMAIL)) {
                    if (data instanceof String) {
                        // regex for e-mail: source: http://www.regular-expressions.info/regexbuddy/email.html
                        if (((String) data).matches("\\b[A-Za-z0-9._%-]+@[A-Za-z0-9.-]+\\.[A-Za-z]{2,4}\\b")) {
                            return true;
                        }
                    }
                } else if (fmt.equals(JSONValueFormat.IP_ADDRESS)) {
                    if (data instanceof String) {
                        // TODO: choose: regular expression or manually parsed version.
                        // IP match using regular expression, source: http://www.regular-expressions.info/regexbuddy/ipaccurate.html
                        boolean isIP = ((String) data).matches("\\b(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\b");
                        if (isIP) {
                            return true;
                        }
//                        if (isIP) {
//                            //return true;
//                        }
//                        String[] ip = ((String) data).split("\\.");
//                        if (ip.length == 4) {
//                            boolean OK = true;
//                            for (String s : ip) {
//                                short sh = Short.parseShort(s);
//                                if (sh < 0 || sh > 255) {
//                                    OK = false;
//                                    break;
//                                }
//                            }
//                            //System.out.println("IP "+data+", isIP="+isIP+", "+OK);
//                            if (OK) {
//                                return true;
//                            }
//                        }
                    }
                } else if (fmt.equals(JSONValueFormat.IPV6)) {
                    // test for IPv6 conformance...
                    if (data instanceof String) {
                        boolean OK = true;
                        String[] ip = ((String) data).split(":");
                        Integer[] ip6 = new Integer[8];
                        int zeroIdx = -1;
                        for (int i = 0; i < ip.length; i++) {
                            if ("".equals(ip[i])) {
                                ip6[i] = 0;
                                zeroIdx = i;
                                break;
                            }
                            ip6[i] = Integer.parseInt(ip[i], 16);
                        }
                        if (zeroIdx != -1) {
                            int shift = ip6.length - ip.length;
                            for (int i = ip6.length - 1; i > zeroIdx; i--) {
                                if ("".equals(ip[i - shift])) {
                                    ip6[i] = 0;
                                    continue;
                                }
                                ip6[i] = Integer.parseInt(ip[i - shift], 16);
                            }
                        }
                        for (int i = 0; i < ip6.length; i++) {
                            if (ip6[i] == null) {
                                OK = false;
                            } else if (ip6[i] < 0 || ip6[i] > 65535) {
                                OK = false;
                            }
                        }
                        if (OK) {
                            return true;
                        }
                    }
                } else if (fmt.equals(JSONValueFormat.HOST_NAME)) {
                    // test for valid host name as described at: http://en.wikipedia.org/wiki/Hostname (for http://tools.ietf.org/html/rfc1123)
                    if (data instanceof String) {
                        if (((String) data).length() < 256
                                && ((String) data).matches("[a-zA-Z0-9][a-zA-Z0-9\\-]{0,62}(\\.[a-zA-Z0-9][a-zA-Z0-9\\-]{0,62})*")) {
                            return true;
                        }
                    }
                } else if (fmt.isSchema()) {
                    if (fmt.getSchema().validate(ctx, data, true)) {
                        return true;
                    }
                }
            } catch (ParseException pex) {
                // caused parsing dates
                ctx.addValidationError(new JSONValidationError(ctx.getPath(), data, schema, getAttribute(), "Format error [" + fmt + "] for " + data + ". ERROR: " + pex + "."));
            } catch (NumberFormatException nfex) {
                // caused parsing numbers
                ctx.addValidationError(new JSONValidationError(ctx.getPath(), data, schema, getAttribute(), "Format error [" + fmt + "] for " + data + ". ERROR: " + nfex + "."));
            } catch (PatternSyntaxException psex) {
                // caused parsing regular expressions
                ctx.addValidationError(new JSONValidationError(ctx.getPath(), data, schema, getAttribute(), "Format error [" + fmt + "] for " + data + ". ERROR: " + psex + "."));
            } catch (IllegalArgumentException iaex) {
                // caused trying to evaluate unknown color from predefined colors enum
                ctx.addValidationError(new JSONValidationError(ctx.getPath(), data, schema, getAttribute(), "Format error [" + fmt + "] for " + data + ". ERROR: " + iaex + "."));
            }
            ctx.addValidationError(new JSONValidationError(ctx.getPath(), data, schema, getAttribute(), "Unrecognized or invalid format [" + fmt + "] for " + data + "."));
            return false;
        }
    }

    public static class AV_Extends extends JSONAttributeValidator {

        @Override
        public ATTRIBUTE getAttribute() {
            return ATTRIBUTE._extends;
        }

        @Override
        public boolean validate(JSONValidationContext ctx, Object data, JSchema schema) throws JSONSchemaException {
            if (ctx == null || schema == null) {
                throw new JSONSchemaException("Invalid context or schema: neither can be null.");
            }
            // we ignore min length for non-string data.
            if (!schema.exists(getAttribute())) {
                return true;
            }

            boolean result = true;
            List<JSONValueExtends> exts = schema.getExtends();
            for (JSONValueExtends ext : exts) {
                if (ext.isSchema()) {
                    if (!ext.getSchema().validate(ctx, data, true)) {
                        result = false;
                    }
                }
            }
            return result;
        }
    }

    /**
     * Tests structure and properties.
     */
    public static class ObjectValidator extends JSONDataValidator {

        public boolean validate(JSONValidationContext ctx, Object data, JSchema schema, boolean recursive) throws JSONSchemaException {
            if (data instanceof Map || (ctx.getReflectiveBuilder() != null && ctx.getReflectiveBuilder().testIsReflectable(data))) {
                Path path = ctx.getPath();
                boolean ok = true;
                Map map = (data instanceof Map) ? (Map) data : ctx.getReflectiveBuilder().getReflectable(data);
                Set<String> processed = new HashSet<String>();
                // validate properties of Object -> subSchema
                for (Object oKey : map.keySet()) {
                    String key = (String) oKey;
                    Object value = map.get(key);
                    JSchema subSchema = schema.getProperty(key);
                    if (subSchema != null) {
                        processed.add(key);
                        path.add(key);
                        if (!subSchema.validate(ctx, value, recursive)) {
                            ok = false;
                        }
                        path.remove();
                    }
                }
                // validate unresolved object properties subSchema -> object
                if (schema.getProperties() != null) {
                    for (String key : schema.getProperties().keySet()) {
                        JSchema subSchema = schema.getProperty(key);
                        path.add(key);

                        if (!processed.contains(key) && subSchema.hasDefault() && ctx.isAssignMissingDefaults()) {
                            try {
                                Object value = subSchema.getDefault();
                                if (data instanceof Map) {
                                    ((Map) data).put(key, value);
                                } else {
                                    ctx.getReflectiveBuilder().setReflectable(data, key, value);
                                }
                                //System.out.println("Assigned to " + path + " default " + value);
                            } catch (Throwable th) {
                                throw new JSONSchemaException("Failed to assign default value at " + path + ": " + th, th);
                            }
                        } else if (!processed.contains(key) && !subSchema.validate(ctx, map.get(key), recursive)) {
                            ok = false;
                        }
                        path.remove();
                    }
                }
                return ok;
            } else {
                // NOT object...
                return false;
            }
        }

        public Iterable<ATTRIBUTE> getTestedAttributes() {
            List<ATTRIBUTE> l = new ArrayList<ATTRIBUTE>(2);
            l.add(ATTRIBUTE._properties);
            l.add(ATTRIBUTE._patternProperties);
            l.add(ATTRIBUTE._default);
            return l;
        }
    }

    /**
     * Tests structure and listed (nested) items.
     */
    public static class ListValidator extends JSONDataValidator {

        public boolean validate(JSONValidationContext ctx, Object data, JSchema schema, boolean recursive) throws JSONSchemaException {
            if (data instanceof Iterable) {
                List<JSONValueType> types = schema.getType();
                if (types.contains(JSONValueType.ARRAY) || types.contains(JSONValueType.ANY)) {
                    Path path = ctx.getPath();
                    boolean ok = true;
                    int idx = 0;
                    JSchema itemSchema = schema.getItems();
                    if (itemSchema == null) {
                        ctx.addValidationError(new JSONValidationError(path, data, schema, ATTRIBUTE._items, "Missing items definition for array.", false));
                    } else {
                        // test min/max items
                        JSONAttributeValidator av = ctx.getValidator(ATTRIBUTE._minItems);
                        if (av != null) {
                            if (!av.validate(ctx, data, itemSchema)) {
                                ok = false;
                            }
                        }
                        av = ctx.getValidator(ATTRIBUTE._maxItems);
                        if (av != null) {
                            if (!av.validate(ctx, data, itemSchema)) {
                                ok = false;
                            }
                        }
                        av = ctx.getValidator(ATTRIBUTE._uniqueItems);
                        if (av != null) {
                            if (!av.validate(ctx, data, itemSchema)) {
                                ok = false;
                            }
                        }
                        // verify nested objects
                        for (Object item : (Iterable) data) {
                            path.setIndex(idx++, true);
                            if (!itemSchema.validate(ctx, item, recursive)) {
                                ok = false;
                            }
                        }
                    }
                    path.setIndex(null, true);
                    return ok;
                } else {
                    ctx.addValidationError(new JSONValidationError(ctx.getPath(), data, schema, ATTRIBUTE._type, "Data type is array [" + data.getClass().getName() + "], expected " + schema.getType()));
                    return false;
                }
            } else {
                // not list/array
                return false;
            }
        }

        public Iterable<ATTRIBUTE> getTestedAttributes() {
            List<ATTRIBUTE> l = new ArrayList<ATTRIBUTE>(3);
            l.add(ATTRIBUTE._items);
            l.add(ATTRIBUTE._minItems);
            l.add(ATTRIBUTE._maxItems);
            l.add(ATTRIBUTE._uniqueItems);
            return l;
        }
    }

    /**
     * Verifies schema attributes are defined correctly (e.g. minimum is <=maximum etc.)
     */
    public static class SchemaValidator extends JSONDataValidator {

        public boolean validate(JSONValidationContext ctx, Object data, JSchema schema, boolean recursive) throws JSONSchemaException {
            // TODO: verify schema element...
            return true;
        }

        public Iterable<ATTRIBUTE> getTestedAttributes() {
            List<ATTRIBUTE> l = new ArrayList<ATTRIBUTE>(16);
            l.add(ATTRIBUTE._title);
            l.add(ATTRIBUTE._description);
            l.add(ATTRIBUTE._id);
            l.add(ATTRIBUTE.$ref);
            l.add(ATTRIBUTE.$schema);
            l.add(ATTRIBUTE._links);
            l.add(ATTRIBUTE._href);
            l.add(ATTRIBUTE._rel);
            l.add(ATTRIBUTE._targetSchema);
            l.add(ATTRIBUTE._method);
            l.add(ATTRIBUTE._enctype);
            l.add(ATTRIBUTE._schema);
            l.add(ATTRIBUTE._fragmentResolution);
            l.add(ATTRIBUTE._readonly);
            l.add(ATTRIBUTE._contentEncoding);
            l.add(ATTRIBUTE._pathStart);
            l.add(ATTRIBUTE._mediaType);
            return l;
        }
    }
}
