/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml.xml;

import com.ssg.tools.jsonxml.Parser;
import com.ssg.tools.jsonxml.common.ParserContext;
import com.ssg.tools.jsonxml.common.Formats;
import com.ssg.tools.jsonxml.common.StructureProcessingException;
import com.ssg.tools.jsonxml.common.Path;
import com.ssg.tools.jsonxml.common.tools.URIResolverExtended;
import java.io.IOException;
import java.io.Reader;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.transform.TransformerFactory;
import org.xml.sax.Attributes;
import org.xml.sax.DTDHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;

/**
 * Generic default SAX-based XML parser.
 * Returns objects structure similar to JSON parser.
 *
 * @author ssg
 */
public class XMLParser extends Parser {

    static Logger log = Logger.getLogger(XMLParser.class.getName());

    public boolean elementIsList(DefaultHandler handler, String uri, String localName, String qName, Attributes attributes) {
        return "list".equals(attributes.getValue("json"));
    }

    public Object preProcessElementObject(XMLParserContext context, Object object) {
        Path path = context.getPath();
        return object;
    }

    /**
     * Converts MAP for single unnamed item into item itself.
     * Additionally checks item for primitive type and replaces if true.
     *
     * @param path
     * @param object
     * @return
     */
    public Object postProcessElementObject(XMLParserContext context, Object object) {
        if (object instanceof Map && ((Map) object).containsKey("")) {
            if (((Map) object).size() == 1) {
                // map -> value
                object = ((Map) object).get("");
            } else if (((Map) object).get("").equals("")) {
                ((Map) object).remove("");
            }
        }
        Object primitive = context.getFormats().toPrimitiveValue(object);
        if (primitive != null) {
            object = primitive;
        }
        Path path = context.getPath();

        log.log(Level.FINEST, "XMLParser: postProcess: " + path + " [" + ((object != null) ? object.getClass().getSimpleName() : "null") + "] = " + object);
        return object;
    }

    /**
     * Used when parent is neither Map nor List instance.
     * By default no action (value is lost!).
     * 
     * @param context
     * @param name
     * @param object 
     */
    public void postProcessReflectiveObject(XMLParserContext context, String name, Object object, int index) throws StructureProcessingException {
    }

    public Object parse(Reader r) throws IOException {
        return parse(r, (Formats) null);
    }

    public Object parse(Reader r, Formats formatters) throws IOException, StructureProcessingException {
        final XMLParserContext context = new XMLParserContext();
        if (formatters != null) {
            context.setFormats(formatters);
        }
        return parse(r, context);
    }

    public Object parse(Reader r, ParserContext xmlContext) throws IOException, StructureProcessingException {
        final XMLParserContext context = (xmlContext != null)
                ? (xmlContext instanceof XMLParserContext)
                ? (XMLParserContext) xmlContext
                : new XMLParserContext(xmlContext)
                : new XMLParserContext();

        try {
            URIResolverExtended resolver = null;
            if (!(TransformerFactory.newInstance().getURIResolver() instanceof URIResolverExtended)) {
                resolver = new URIResolverExtended();
                TransformerFactory.newInstance().setURIResolver(resolver);
            } else {
                resolver = (URIResolverExtended) TransformerFactory.newInstance().getURIResolver();
            }

            SAXParserFactory factory = SAXParserFactory.newInstance();
            factory.setValidating(false);
            SAXParser saxParser = factory.newSAXParser();
            saxParser.getXMLReader().setEntityResolver(resolver);
            XMLReader rr = saxParser.getXMLReader();
            rr.setEntityResolver(resolver);
            rr.setDTDHandler(new DTDHandler() {

                public void notationDecl(String name, String publicId, String systemId) throws SAXException {
                    //System.out.println(getClass().getName() + ".notationDecl(" + name + ", " + publicId + ", " + systemId + ")");
                }

                public void unparsedEntityDecl(String name, String publicId, String systemId, String notationName) throws SAXException {
                    //System.out.println(getClass().getName() + ".unparsedEntityDecl(" + name + ", " + publicId + ", " + systemId + ", " + notationName + ")");
                }
            });
            rr.setContentHandler(new DefaultHandler() {

                @Override
                public void startDocument() throws SAXException {
                    context.getStack().push(new LinkedHashMap());
                }

                @Override
                public void endDocument() throws SAXException {
                    // TODO: ensure resultant object is good...
                }

                @Override
                public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException {
                    if (context.getStack().peek() instanceof Map && ((Map) context.getStack().peek()).containsKey(qName) && ((Map) context.getStack().peek()).size() == 1) {
                        // convert current Map into list...
                        //System.out.println("startElement: convert parent from Map to List: " + qName + ", " + context.getPath());
                        Map m = (Map) context.getStack().pop();
                        List l = new ArrayList();
                        context.getStack().push(l);
                        l.add(m.get(qName));
                        context.getPath().remove();
                        context.getPath().add(qName, 0);
                    }
                    if (elementIsList(this, uri, localName, qName, attributes)) {
                        //System.out.println("startElement: added List : " + qName);
                        context.getStack().push(new ArrayList());
                        context.getPath().add(qName, 0);
                    } else {
                        //System.out.println("startElement: added Map  : " + qName);
                        Map map = new LinkedHashMap();
                        context.getStack().push(map);
                        context.getPath().add(qName);
                        for (int i = 0; i < attributes.getLength(); i++) {
                            String name = attributes.getQName(i).trim();
                            String value = attributes.getValue(i);
                            if (name != null && name.trim().length() != 0) {
                                map.put(context.getAttributePrefix() + name, value);
                            }
                        }
                    }
                    Object preProcessed = preProcessElementObject(context, context.getStack().peek());
                    if (preProcessed != null && preProcessed.equals(context.getStack().peek())) {
                        // replace original with pre-processed
                        context.getStack().pop();
                        context.getStack().push(preProcessed);
                    }
                }

                @Override
                public void characters(char[] ch, int start, int length) throws SAXException {
                    String value = new String(ch, start, length);
                    value = value.trim();
                    if (context.getStack().peek() instanceof Map) {
                        Map m = (Map) context.getStack().peek();
                        if (m.containsKey("")) {
                            m.put("", m.get("") + value);
                        } else {
                            m.put("", value);
                        }
                    }
                }

                @Override
                public void endElement(String uri, String localName, String qName) throws SAXException {
                    Object obj = context.getStack().pop();
                    context.getPath().remove();
                    //System.out.println("endElement: " + obj);
                    obj = postProcessElementObject(context, obj);
                    if (context.getStack().peek() instanceof List) {
                        ((List) context.getStack().peek()).add(obj);
                        String last = context.getPath().remove();
                        context.getPath().add(last, ((List) context.getStack().peek()).size());
                        //System.out.println("endElement: Item is added to list. Size=" + ((List) context.getStack().peek()).size());
                    } else if (context.getStack().peek() instanceof Map) {
                        Map parent = (Map) context.getStack().peek();
                        if (parent.containsKey(qName)) {
                            if (!(parent.get(qName) instanceof List)) {
                                //System.out.println("endElement: convert parent from Map to List: " + qName + ", " + context.getPath());
                                Object first = parent.get(qName);
                                List list = new ArrayList();
                                list.add(first);
                                parent.put(qName, list);
                                if (parent.size() == 2) {
                                    context.getStack().pop();
                                    context.getStack().push(list);
                                }
                            }
                            ((List) parent.get(qName)).add(obj);

                        } else {
                            parent.put(qName, obj);
                        }
                        //System.out.println("Item is added to map as "+qName+". Size="+((Map)stack.peek()).size());
                        //} else {
                    }

                    {
                        //System.err.println("Item is lost (no container: " + stack.peek().getClass().getName() + "): " + qName + ": " + obj);
                        try {
                            postProcessReflectiveObject(context, qName, obj, (context.getStack().peek() instanceof List) ? ((List) context.getStack().peek()).size() : -1);
                        } catch (StructureProcessingException pex) {
                            throw new SAXException("Nested error.", pex);
                        }
                    }
                }
            });
            rr.parse(new InputSource(r));
        } catch (ParserConfigurationException pcex) {
            throw new IOException("SAXParser configuration error: " + pcex, pcex);
        } catch (SAXException sex) {
            throw new IOException("SAXParser parsing error: " + sex, sex);
        }
        return context.getStack().pop();
    }

    /**
     * Method for testing source as syntactically valid XML.
     * In case of problems SAXException is thrown.
     * This method uses same entity and DTD resolution as parser and just misses creation of any object.
     * 
     * @param r
     * @throws IOException
     * @throws SAXException 
     */
    public static void test(Reader r) throws IOException, SAXException {
        try {
            URIResolverExtended resolver = null;
            if (!(TransformerFactory.newInstance().getURIResolver() instanceof URIResolverExtended)) {
                resolver = new URIResolverExtended();
                TransformerFactory.newInstance().setURIResolver(resolver);
            } else {
                resolver = (URIResolverExtended) TransformerFactory.newInstance().getURIResolver();
            }

            SAXParserFactory factory = SAXParserFactory.newInstance();
            factory.setValidating(false);
            SAXParser saxParser = factory.newSAXParser();
            saxParser.getXMLReader().setEntityResolver(resolver);
            XMLReader rr = saxParser.getXMLReader();
            rr.setEntityResolver(resolver);
            rr.setDTDHandler(new DTDHandler() {

                public void notationDecl(String name, String publicId, String systemId) throws SAXException {
                }

                public void unparsedEntityDecl(String name, String publicId, String systemId, String notationName) throws SAXException {
                }
            });
            rr.setContentHandler(new DefaultHandler() {

                @Override
                public void startDocument() throws SAXException {
                }

                @Override
                public void endDocument() throws SAXException {
                }

                @Override
                public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException {
                }

                @Override
                public void characters(char[] ch, int start, int length) throws SAXException {
                }

                @Override
                public void endElement(String uri, String localName, String qName) throws SAXException {
                }
            });
            rr.parse(new InputSource(r));
        } catch (ParserConfigurationException pcex) {
            throw new IOException("SAXParser configuration error: " + pcex, pcex);
        } catch (SAXException sex) {
            throw new IOException("SAXParser parsing error: " + sex, sex);
        }
    }
}
